import abc
import json
import threading
import time
from typing import Any, Callable, Optional, Set, Union

from moto.stepfunctions.parser.api import (
    HistoryEventExecutionDataDetails,
    HistoryEventType,
    TaskFailedEventDetails,
    TaskSubmittedEventDetails,
)
from moto.stepfunctions.parser.asl.component.common.error_name.custom_error_name import (
    CustomErrorName,
)
from moto.stepfunctions.parser.asl.component.common.error_name.failure_event import (
    FailureEvent,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_task.credentials import (
    ComputedCredentials,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_task.service.resource import (
    ResourceCondition,
    ResourceRuntimePart,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_task.service.state_task_service import (
    StateTaskService,
)
from moto.stepfunctions.parser.asl.eval.callback.callback import (
    CallbackEndpoint,
    CallbackOutcome,
    CallbackOutcomeFailure,
    CallbackOutcomeFailureError,
    CallbackOutcomeSuccess,
    CallbackOutcomeTimedOut,
    CallbackTimeoutError,
    HeartbeatEndpoint,
    HeartbeatTimedOut,
    HeartbeatTimeoutError,
)
from moto.stepfunctions.parser.asl.eval.environment import Environment
from moto.stepfunctions.parser.asl.eval.event.event_detail import EventDetails
from moto.stepfunctions.parser.asl.utils.encoding import to_json_str
from moto.stepfunctions.parser.utils import TMP_THREADS

# TODO: consider implementing a polling pattern similar to that observable from AWS:
# https://repost.aws/questions/QUFFlHcbvIQFe-bS3RAi7TWA/a-glue-job-in-a-step-function-is-taking-so-long-to-continue-the-next-step
_DELAY_SECONDS_SYNC_CONDITION_CHECK: float = 0.5


class StateTaskServiceCallback(StateTaskService, abc.ABC):
    _supported_integration_patterns: Set[ResourceCondition]

    def __init__(self, supported_integration_patterns: Set[ResourceCondition]):
        super().__init__()
        self._supported_integration_patterns = supported_integration_patterns

    def _get_sfn_resource(self) -> str:
        resource = super()._get_sfn_resource()
        if self.resource.condition is not None:
            resource += f".{self.resource.condition}"
        return resource

    def _build_sync_resolver(
        self,
        env: Environment,
        resource_runtime_part: ResourceRuntimePart,
        normalised_parameters: dict,
        task_credentials: ComputedCredentials,
    ) -> Callable[[], Optional[Any]]:
        raise RuntimeError(
            f"Unsupported .sync callback procedure in resource {self.resource.resource_arn}"
        )

    def _build_sync2_resolver(
        self,
        env: Environment,
        resource_runtime_part: ResourceRuntimePart,
        normalised_parameters: dict,
        task_credentials: ComputedCredentials,
    ) -> Callable[[], Optional[Any]]:
        raise RuntimeError(
            f"Unsupported .sync2 callback procedure in resource {self.resource.resource_arn}"
        )

    def _eval_wait_for_task_token(
        self,
        env: Environment,
        timeout_seconds: int,
        callback_endpoint: CallbackEndpoint,
        heartbeat_endpoint: Optional[HeartbeatEndpoint],
    ) -> CallbackOutcome:
        outcome: Optional[CallbackOutcome]
        if heartbeat_endpoint is not None:
            outcome = self._wait_for_task_token_heartbeat(
                env, callback_endpoint, heartbeat_endpoint
            )
        else:
            outcome = self._wait_for_task_token_timeout(
                timeout_seconds, callback_endpoint
            )
        if outcome is None:
            return CallbackOutcomeTimedOut(callback_id=callback_endpoint.callback_id)
        return outcome

    def _eval_sync(
        self,
        env: Environment,
        sync_resolver: Callable[[], Optional[Any]],
        timeout_seconds: Optional[int],
        callback_endpoint: Optional[CallbackEndpoint],
        heartbeat_endpoint: Optional[HeartbeatEndpoint],
    ) -> Union[CallbackOutcome, Any]:
        callback_output: Optional[CallbackOutcome] = None

        # Listen for WaitForTaskToken signals if an endpoint is provided.
        if callback_endpoint is not None:

            def _local_update_wait_for_task_token():
                nonlocal callback_output
                callback_output = self._eval_wait_for_task_token(
                    env=env,
                    timeout_seconds=timeout_seconds,
                    callback_endpoint=callback_endpoint,
                    heartbeat_endpoint=heartbeat_endpoint,
                )

            thread_wait_for_task_token = threading.Thread(
                target=_local_update_wait_for_task_token,
                name=f"WaitForTaskToken_SyncTask_{self.resource.resource_arn}",
                daemon=True,
            )
            TMP_THREADS.append(thread_wait_for_task_token)
            thread_wait_for_task_token.start()
            # Note: the stopping of this worker thread is handled indirectly through the state of env.
            #       an exception in this thread will invalidate env, and therefore the worker thread.
            #       hence why here there are no explicit stopping logic for thread_wait_for_task_token.

        sync_result: Optional[Any] = None
        while env.is_running():
            sync_result = sync_resolver()
            if callback_output or sync_result:
                break
            else:
                time.sleep(_DELAY_SECONDS_SYNC_CONDITION_CHECK)

        return callback_output or sync_result

    def _eval_integration_pattern(
        self,
        env: Environment,
        resource_runtime_part: ResourceRuntimePart,
        normalised_parameters: dict,
        task_credentials: ComputedCredentials,
    ) -> None:
        task_output = env.stack.pop()

        # Initialise the waitForTaskToken Callback endpoint for this task if supported.
        callback_endpoint: Optional[CallbackEndpoint] = None
        if ResourceCondition.WaitForTaskToken in self._supported_integration_patterns:
            callback_id = env.states.context_object.context_object_data["Task"]["Token"]
            callback_endpoint = env.callback_pool_manager.get(callback_id)

        # Setup resources for timeout control.
        self.timeout.eval(env=env)
        timeout_seconds = env.stack.pop()

        # Setup resources for heartbeat workloads if necessary.
        heartbeat_endpoint: Optional[HeartbeatEndpoint] = None
        if self.heartbeat:
            self.heartbeat.eval(env=env)
            heartbeat_seconds = env.stack.pop()
            heartbeat_endpoint: HeartbeatEndpoint = (
                callback_endpoint.setup_heartbeat_endpoint(
                    heartbeat_seconds=heartbeat_seconds
                )
            )

        # Collect the output of the integration pattern.
        outcome: Union[CallbackOutcome, Any]
        try:
            if self.resource.condition == ResourceCondition.WaitForTaskToken:
                outcome = self._eval_wait_for_task_token(
                    env=env,
                    timeout_seconds=timeout_seconds,
                    callback_endpoint=callback_endpoint,
                    heartbeat_endpoint=heartbeat_endpoint,
                )
            else:
                # Sync operations require the task output as input.
                env.stack.append(task_output)
                if self.resource.condition == ResourceCondition.Sync:
                    sync_resolver = self._build_sync_resolver(
                        env=env,
                        resource_runtime_part=resource_runtime_part,
                        normalised_parameters=normalised_parameters,
                        task_credentials=task_credentials,
                    )
                else:
                    # The condition checks about the resource's condition is exhaustive leaving
                    # only Sync2 ResourceCondition types in this block.
                    sync_resolver = self._build_sync2_resolver(
                        env=env,
                        resource_runtime_part=resource_runtime_part,
                        normalised_parameters=normalised_parameters,
                        task_credentials=task_credentials,
                    )

                outcome = self._eval_sync(
                    env=env,
                    timeout_seconds=timeout_seconds,
                    callback_endpoint=callback_endpoint,
                    heartbeat_endpoint=heartbeat_endpoint,
                    sync_resolver=sync_resolver,
                )
        except Exception as integration_exception:
            outcome = integration_exception
        finally:
            # Now that the outcome is collected or the exception is about to be passed upstream, and the process has
            # finished, ensure all waiting # threads on this endpoint (or task) will stop. This is in an effort to
            # release resources sooner than when these would eventually synchronise with the updated environment
            # state of this task.
            callback_endpoint.interrupt_all()

        # Handle Callback outcome types.
        if isinstance(outcome, CallbackOutcomeTimedOut):
            raise CallbackTimeoutError()
        elif isinstance(outcome, HeartbeatTimedOut):
            raise HeartbeatTimeoutError()
        elif isinstance(outcome, CallbackOutcomeFailure):
            raise CallbackOutcomeFailureError(callback_outcome_failure=outcome)
        elif isinstance(outcome, CallbackOutcomeSuccess):
            outcome_output = json.loads(outcome.output)
            env.stack.append(outcome_output)
        # Pass evaluation exception upstream for error handling.
        elif isinstance(outcome, Exception):
            raise outcome
        # Otherwise the outcome is the result of the integration pattern (sync, sync2)
        # therefore push it onto the evaluation stack for the next operations.
        else:
            env.stack.append(outcome)

    def _wait_for_task_token_timeout(  # noqa
        self,
        timeout_seconds: int,
        callback_endpoint: CallbackEndpoint,
    ) -> Optional[CallbackOutcome]:
        # Awaits a callback notification and returns the outcome received.
        # If the operation times out or is interrupted it returns None.

        # Although the timeout is handled already be the superclass (ExecutionState),
        # the timeout value is specified here too, to allow this child process to terminate earlier even if
        # discarded by the main process.
        # Note: although this is the same timeout value, this can only decay strictly after the first timeout
        #       started as it is invoked strictly later.
        outcome: Optional[CallbackOutcome] = callback_endpoint.wait(
            timeout=timeout_seconds
        )
        return outcome

    def _wait_for_task_token_heartbeat(  # noqa
        self,
        env: Environment,
        callback_endpoint: CallbackEndpoint,
        heartbeat_endpoint: HeartbeatEndpoint,
    ) -> Optional[CallbackOutcome]:
        outcome = None
        while (
            env.is_running()
            and outcome
            is None  # Note: the lifetime of this environment is this task's not the entire state machine program.
        ):  # Until subprocess hasn't timed out or result wasn't received.
            received = heartbeat_endpoint.clear_and_wait()
            if not received and env.is_running():  # Heartbeat timed out.
                outcome = HeartbeatTimedOut()
            else:
                outcome = callback_endpoint.get_outcome()
        return outcome

    def _assert_integration_pattern_is_supported(self):
        integration_pattern = self.resource.condition
        if integration_pattern not in self._supported_integration_patterns:
            raise RuntimeError(
                f"Unsupported {integration_pattern} callback procedure in resource {self.resource.resource_arn}"
            )

    def _is_integration_pattern(self):
        return self.resource.condition is not None

    def _get_callback_outcome_failure_event(
        self, env: Environment, ex: CallbackOutcomeFailureError
    ) -> FailureEvent:
        callback_outcome_failure: CallbackOutcomeFailure = ex.callback_outcome_failure
        error: Optional[str] = callback_outcome_failure.error
        return FailureEvent(
            env=env,
            error_name=CustomErrorName(error_name=error),
            event_type=HistoryEventType.TaskFailed,
            event_details=EventDetails(
                taskFailedEventDetails=TaskFailedEventDetails(
                    resourceType=self._get_sfn_resource_type(),
                    resource=self._get_sfn_resource(),
                    error=error,
                    cause=callback_outcome_failure.cause,
                )
            ),
        )

    def _from_error(self, env: Environment, ex: Exception) -> FailureEvent:
        if isinstance(ex, CallbackOutcomeFailureError):
            return self._get_callback_outcome_failure_event(env=env, ex=ex)
        return super()._from_error(env=env, ex=ex)

    def _eval_body(self, env: Environment) -> None:
        # Generate a TaskToken uuid within the context object, if this task resources has a callback condition.
        # https://docs.aws.amazon.com/step-functions/latest/dg/connect-to-resource.html#connect-wait-token
        if (
            self._is_integration_pattern()
            and ResourceCondition.WaitForTaskToken
            in self._supported_integration_patterns
        ):
            self._assert_integration_pattern_is_supported()
            task_token = env.states.context_object.update_task_token()
            env.callback_pool_manager.add(task_token)

        super()._eval_body(env=env)

        # Ensure the TaskToken field is reset, as this is only available during waitForTaskToken task evaluations.
        env.states.context_object.context_object_data.pop("Task", None)

    def _after_eval_execution(
        self,
        env: Environment,
        resource_runtime_part: ResourceRuntimePart,
        normalised_parameters: dict,
        task_credentials: ComputedCredentials,
    ) -> None:
        if self._is_integration_pattern():
            output = env.stack[-1]
            env.event_manager.add_event(
                context=env.event_history_context,
                event_type=HistoryEventType.TaskSubmitted,
                event_details=EventDetails(
                    taskSubmittedEventDetails=TaskSubmittedEventDetails(
                        resource=self._get_sfn_resource(),
                        resourceType=self._get_sfn_resource_type(),
                        output=to_json_str(output),
                        outputDetails=HistoryEventExecutionDataDetails(truncated=False),
                    )
                ),
            )
            self._eval_integration_pattern(
                env=env,
                resource_runtime_part=resource_runtime_part,
                normalised_parameters=normalised_parameters,
                task_credentials=task_credentials,
            )

        super()._after_eval_execution(
            env=env,
            resource_runtime_part=resource_runtime_part,
            normalised_parameters=normalised_parameters,
            task_credentials=task_credentials,
        )



from moto.utilities.tagging_service import TaggingService


def test_list_empty():
    svc = TaggingService()
    result = svc.list_tags_for_resource("test")

    assert {"Tags": []} == result


def test_create_tag():
    svc = TaggingService("TheTags", "TagKey", "TagValue")
    tags = [{"TagKey": "key_key", "TagValue": "value_value"}]
    svc.tag_resource("arn", tags)
    actual = svc.list_tags_for_resource("arn")
    expected = {"TheTags": [{"TagKey": "key_key", "TagValue": "value_value"}]}

    assert expected == actual


def test_create_tag_without_value():
    svc = TaggingService()
    tags = [{"Key": "key_key"}]
    svc.tag_resource("arn", tags)
    actual = svc.list_tags_for_resource("arn")
    expected = {"Tags": [{"Key": "key_key", "Value": None}]}

    assert expected == actual


def test_delete_tag_using_names():
    svc = TaggingService()
    tags = [{"Key": "key_key", "Value": "value_value"}]
    svc.tag_resource("arn", tags)
    svc.untag_resource_using_names("arn", ["key_key"])
    result = svc.list_tags_for_resource("arn")

    assert {"Tags": []} == result


def test_delete_all_tags_for_resource():
    svc = TaggingService()
    tags = [{"Key": "key_key", "Value": "value_value"}]
    tags2 = [{"Key": "key_key2", "Value": "value_value2"}]
    svc.tag_resource("arn", tags)
    svc.tag_resource("arn", tags2)
    svc.delete_all_tags_for_resource("arn")
    result = svc.list_tags_for_resource("arn")

    assert {"Tags": []} == result


def test_list_empty_delete():
    svc = TaggingService()
    svc.untag_resource_using_names("arn", ["key_key"])
    result = svc.list_tags_for_resource("arn")

    assert {"Tags": []} == result


def test_delete_tag_using_tags():
    svc = TaggingService()
    tags = [{"Key": "key_key", "Value": "value_value"}]
    svc.tag_resource("arn", tags)
    svc.untag_resource_using_tags("arn", tags)
    result = svc.list_tags_for_resource("arn")

    assert {"Tags": []} == result


def test_extract_tag_names():
    svc = TaggingService()
    tags = [{"Key": "key1", "Value": "value1"}, {"Key": "key2", "Value": "value2"}]
    actual = svc.extract_tag_names(tags)
    expected = ["key1", "key2"]

    assert expected == actual


def test_copy_non_existing_arn():
    svc = TaggingService()
    tags = [{"Key": "key1", "Value": "value1"}, {"Key": "key2", "Value": "value2"}]
    svc.tag_resource("new_arn", tags)
    #
    svc.copy_tags("non_existing_arn", "new_arn")
    # Copying from a non-existing ARN should a NOOP
    # Assert the old tags still exist
    actual = sorted(
        svc.list_tags_for_resource("new_arn")["Tags"], key=lambda t: t["Key"]
    )
    assert actual == tags


def test_copy_existing_arn():
    svc = TaggingService()
    tags_old_arn = [{"Key": "key1", "Value": "value1"}]
    tags_new_arn = [{"Key": "key2", "Value": "value2"}]
    svc.tag_resource("old_arn", tags_old_arn)
    svc.tag_resource("new_arn", tags_new_arn)
    #
    svc.copy_tags("old_arn", "new_arn")
    # Assert the old tags still exist
    actual = sorted(
        svc.list_tags_for_resource("new_arn")["Tags"], key=lambda t: t["Key"]
    )
    assert actual == [
        {"Key": "key1", "Value": "value1"},
        {"Key": "key2", "Value": "value2"},
    ]


def test_validate_tags():
    """Unit tests for validate_tags()."""
    svc = TaggingService()
    # Key with invalid characters.
    errors = svc.validate_tags([{"Key": "foo!", "Value": "bar"}])
    assert (
        "Value 'foo!' at 'tags.1.member.key' failed to satisfy constraint: "
        "Member must satisfy regular expression pattern"
    ) in errors

    # Value with invalid characters.
    errors = svc.validate_tags([{"Key": "foo", "Value": "bar!"}])
    assert (
        "Value 'bar!' at 'tags.1.member.value' failed to satisfy "
        "constraint: Member must satisfy regular expression pattern"
    ) in errors

    # Key too long.
    errors = svc.validate_tags(
        [
            {
                "Key": (
                    "12345678901234567890123456789012345678901234567890"
                    "12345678901234567890123456789012345678901234567890"
                    "123456789012345678901234567890"
                ),
                "Value": "foo",
            }
        ]
    )
    assert (
        "at 'tags.1.member.key' failed to satisfy constraint: Member must "
        "have length less than or equal to 128"
    ) in errors

    # Value too long.
    errors = svc.validate_tags(
        [
            {
                "Key": "foo",
                "Value": (
                    "12345678901234567890123456789012345678901234567890"
                    "12345678901234567890123456789012345678901234567890"
                    "12345678901234567890123456789012345678901234567890"
                    "12345678901234567890123456789012345678901234567890"
                    "12345678901234567890123456789012345678901234567890"
                    "1234567890"
                ),
            },
        ]
    )
    assert (
        "at 'tags.1.member.value' failed to satisfy constraint: Member "
        "must have length less than or equal to 256"
    ) in errors

    # Compound errors.
    errors = svc.validate_tags(
        [
            {
                "Key": (
                    "12345678901234567890123456789012345678901234567890"
                    "12345678901234567890123456789012345678901234567890"
                    "123456789012345678901234567890"
                ),
                "Value": (
                    "12345678901234567890123456789012345678901234567890"
                    "12345678901234567890123456789012345678901234567890"
                    "12345678901234567890123456789012345678901234567890"
                    "12345678901234567890123456789012345678901234567890"
                    "12345678901234567890123456789012345678901234567890"
                    "1234567890"
                ),
            },
            {"Key": "foo!", "Value": "bar!"},
        ]
    )
    assert "4 validation errors detected" in errors
    assert (
        "at 'tags.1.member.key' failed to satisfy constraint: Member must "
        "have length less than or equal to 128"
    ) in errors
    assert (
        "at 'tags.1.member.value' failed to satisfy constraint: Member "
        "must have length less than or equal to 256"
    ) in errors
    assert (
        "Value 'foo!' at 'tags.2.member.key' failed to satisfy constraint: "
        "Member must satisfy regular expression pattern"
    ) in errors
    assert (
        "Value 'bar!' at 'tags.2.member.value' failed to satisfy "
        "constraint: Member must satisfy regular expression pattern"
    ) in errors



from unittest import SkipTest

import boto3
import pytest
from botocore.exceptions import ClientError

from moto import mock_aws, settings


@pytest.fixture(name="client")
def client_fixture():
    with mock_aws():
        yield boto3.client("servicediscovery", region_name="eu-west-1")


@pytest.fixture(name="ns_resp")
def ns_resp_fixture(client):
    client.create_http_namespace(Name="mynamespace")
    namespace = [
        ns
        for ns in client.list_namespaces()["Namespaces"]
        if ns["Name"] == "mynamespace"
    ][0]
    return dict(Namespace=namespace)


@pytest.fixture(name="srv_resp")
def srv_resp_fixture(client, ns_resp):
    return client.create_service(
        Name="myservice",
        NamespaceId=ns_resp["Namespace"]["Id"],
        DnsConfig={"DnsRecords": [{"Type": "A", "TTL": 60}]},
    )


@mock_aws
def test_register_instance(client, ns_resp, srv_resp):
    instance_id = "i-123"
    creator_request_id = "crid"
    attributes = {"attr1": "value1"}
    inst_resp = client.register_instance(
        ServiceId=srv_resp["Service"]["Id"],
        InstanceId=instance_id,
        CreatorRequestId=creator_request_id,
        Attributes=attributes,
    )

    assert "OperationId" in inst_resp

    operation = client.get_operation(OperationId=inst_resp["OperationId"])
    assert operation["Operation"]["Targets"]["INSTANCE"] == instance_id

    instance = client.get_instance(
        ServiceId=srv_resp["Service"]["Id"], InstanceId=instance_id
    )
    assert instance["Instance"]["CreatorRequestId"] == creator_request_id
    assert instance["Instance"]["Attributes"] == attributes
    assert instance["Instance"]["Id"] == instance_id


@mock_aws
def test_deregister_instance(client, ns_resp, srv_resp):
    instance_id = "i-123"
    creator_request_id = "crid"
    attributes = {"attr1": "value1"}
    client.register_instance(
        ServiceId=srv_resp["Service"]["Id"],
        InstanceId=instance_id,
        CreatorRequestId=creator_request_id,
        Attributes=attributes,
    )

    dereg_resp = client.deregister_instance(
        ServiceId=srv_resp["Service"]["Id"], InstanceId=instance_id
    )
    assert "OperationId" in dereg_resp

    operation = client.get_operation(OperationId=dereg_resp["OperationId"])
    assert operation["Operation"]["Targets"]["INSTANCE"] == instance_id

    with pytest.raises(ClientError) as exc:
        client.get_instance(ServiceId=srv_resp["Service"]["Id"], InstanceId=instance_id)
        assert exc.value.response["Error"]["Code"] == "InstanceNotFound"
        assert exc.value.response["Error"]["Message"] == instance_id


@mock_aws
def test_get_instance(client, ns_resp, srv_resp):
    instance_id = "i-123"
    creator_request_id = "crid"
    attributes = {"attr1": "value1"}
    client.register_instance(
        ServiceId=srv_resp["Service"]["Id"],
        InstanceId=instance_id,
        CreatorRequestId=creator_request_id,
        Attributes=attributes,
    )

    instance = client.get_instance(
        ServiceId=srv_resp["Service"]["Id"], InstanceId=instance_id
    )
    assert "Instance" in instance
    assert instance["Instance"]["CreatorRequestId"] == creator_request_id
    assert instance["Instance"]["Attributes"] == attributes
    assert instance["Instance"]["Id"] == instance_id


@mock_aws
def test_get_unknown_instance(client, srv_resp):
    with pytest.raises(ClientError) as exc:
        client.get_instance(ServiceId=srv_resp["Service"]["Id"], InstanceId="unknown")
    err = exc.value.response["Error"]
    assert err["Code"] == "InstanceNotFound"
    assert err["Message"] == "unknown"


@mock_aws
def test_list_instances(client, ns_resp, srv_resp):
    instance_ids = ["i-123", "i-456", "i-789", "i-012"]
    for instance_id in instance_ids:
        client.register_instance(
            ServiceId=srv_resp["Service"]["Id"], InstanceId=instance_id, Attributes={}
        )

    instances = client.list_instances(ServiceId=srv_resp["Service"]["Id"])
    assert len(instances["Instances"]) == 4
    assert set(inst["Id"] for inst in instances["Instances"]) == set(instance_ids)


@mock_aws
def test_paginate_list_instances(client, ns_resp, srv_resp):
    instance_ids = ["i-123", "i-456", "i-789", "i-012"]
    for instance_id in instance_ids:
        client.register_instance(
            ServiceId=srv_resp["Service"]["Id"], InstanceId=instance_id, Attributes={}
        )

    instances = client.list_instances(ServiceId=srv_resp["Service"]["Id"], MaxResults=2)
    assert len(instances["Instances"]) == 2
    assert "NextToken" in instances
    assert set(inst["Id"] for inst in instances["Instances"]) == set(instance_ids[:2])

    instances = client.list_instances(
        ServiceId=srv_resp["Service"]["Id"], NextToken=instances["NextToken"]
    )
    assert len(instances["Instances"]) == 2
    assert "NextToken" not in instances
    assert set(inst["Id"] for inst in instances["Instances"]) == set(instance_ids[2:])


@mock_aws
def test_get_instances_health_status(client, ns_resp, srv_resp):
    instance_ids = ["i-123", "i-456", "i-789", "i-012"]
    for instance_id in instance_ids:
        client.register_instance(
            ServiceId=srv_resp["Service"]["Id"], InstanceId=instance_id, Attributes={}
        )

    health_status = client.get_instances_health_status(
        ServiceId=srv_resp["Service"]["Id"], Instances=instance_ids
    )
    assert len(health_status["Status"]) == 4
    for inst_id in instance_ids:
        assert health_status["Status"][inst_id] == "HEALTHY"


@mock_aws
def test_update_instance_custom_health_status(client, ns_resp, srv_resp):
    instance_id = "i-123"
    client.register_instance(
        ServiceId=srv_resp["Service"]["Id"], InstanceId=instance_id, Attributes={}
    )

    client.update_instance_custom_health_status(
        ServiceId=srv_resp["Service"]["Id"], InstanceId=instance_id, Status="UNHEALTHY"
    )

    health_status = client.get_instances_health_status(
        ServiceId=srv_resp["Service"]["Id"], Instances=[instance_id]
    )
    assert health_status["Status"][instance_id] == "UNHEALTHY"


@mock_aws
def test_discover_instances_formatting(client, ns_resp, srv_resp):
    if not settings.TEST_DECORATOR_MODE:
        raise SkipTest(
            "Endpoint for discovering instances is prefixed with 'data-', and we can't intercept calls to 'data-localhost'"
        )

    attr_dict = {"attr1": "value1", "attr2": "value1"}
    client.register_instance(
        ServiceId=srv_resp["Service"]["Id"], InstanceId="i-123", Attributes=attr_dict
    )

    instances = client.discover_instances(
        NamespaceName=ns_resp["Namespace"]["Name"],
        ServiceName=srv_resp["Service"]["Name"],
        MaxResults=2,
    )

    assert len(instances["Instances"]) == 1
    assert instances["Instances"][0]["InstanceId"] == "i-123"
    assert instances["Instances"][0]["NamespaceName"] == ns_resp["Namespace"]["Name"]
    assert instances["Instances"][0]["ServiceName"] == srv_resp["Service"]["Name"]
    assert instances["Instances"][0]["Attributes"] == attr_dict
    assert instances["Instances"][0]["HealthStatus"] == "HEALTHY"
    assert instances["InstancesRevision"] == 1


@mock_aws
def test_discover_instances_attr_filters(client, ns_resp, srv_resp):
    if not settings.TEST_DECORATOR_MODE:
        raise SkipTest(
            "Endpoint for discovering instances is prefixed with 'data-', and we can't intercept calls to 'data-localhost'"
        )

    instance_dicts = [
        {"id": "i-123", "attributes": {"attr1": "value1", "attr2": "value1"}},
        {"id": "i-456", "attributes": {"attr1": "value2"}},
        {"id": "i-789", "attributes": {"attr1": "value3", "attr2": "value1"}},
        {"id": "i-012", "attributes": {"attr1": "value3"}},
    ]
    for inst_dict in instance_dicts:
        client.register_instance(
            ServiceId=srv_resp["Service"]["Id"],
            InstanceId=inst_dict["id"],
            Attributes=inst_dict["attributes"],
        )

    instances = client.discover_instances(
        NamespaceName=ns_resp["Namespace"]["Name"],
        ServiceName=srv_resp["Service"]["Name"],
    )
    assert len(instances["Instances"]) == 4
    assert set(inst["InstanceId"] for inst in instances["Instances"]) == {
        "i-123",
        "i-456",
        "i-789",
        "i-012",
    }

    instances = client.discover_instances(
        NamespaceName=ns_resp["Namespace"]["Name"],
        ServiceName=srv_resp["Service"]["Name"],
        QueryParameters={"attr1": "value3"},
    )
    assert len(instances["Instances"]) == 2
    assert set(inst["InstanceId"] for inst in instances["Instances"]) == {
        "i-789",
        "i-012",
    }

    instances = client.discover_instances(
        NamespaceName=ns_resp["Namespace"]["Name"],
        ServiceName=srv_resp["Service"]["Name"],
        QueryParameters={"attr1": "value3"},
        OptionalParameters={"attr2": "value1"},
    )
    assert len(instances["Instances"]) == 1
    assert instances["Instances"][0]["InstanceId"] == "i-789"

    instances = client.discover_instances(
        NamespaceName=ns_resp["Namespace"]["Name"],
        ServiceName=srv_resp["Service"]["Name"],
        QueryParameters={"attr1": "value3"},
        OptionalParameters={"attr2": "value2"},
    )
    assert len(instances["Instances"]) == 2
    assert set(inst["InstanceId"] for inst in instances["Instances"]) == {
        "i-789",
        "i-012",
    }


@mock_aws
def test_discover_instances_health_filters(client, ns_resp, srv_resp):
    if not settings.TEST_DECORATOR_MODE:
        raise SkipTest(
            "Endpoint for discovering instances is prefixed with 'data-', and we can't intercept calls to 'data-localhost'"
        )

    instance_dicts = [
        {"id": "i-123"},
        {"id": "i-456"},
        {"id": "i-789", "health": "UNHEALTHY"},
        {"id": "i-012", "health": "UNHEALTHY"},
    ]
    for inst_dict in instance_dicts:
        client.register_instance(
            ServiceId=srv_resp["Service"]["Id"],
            InstanceId=inst_dict["id"],
            Attributes={},
        )
        if "health" in inst_dict:
            client.update_instance_custom_health_status(
                ServiceId=srv_resp["Service"]["Id"],
                InstanceId=inst_dict["id"],
                Status=inst_dict["health"],
            )

    instances = client.discover_instances(
        NamespaceName=ns_resp["Namespace"]["Name"],
        ServiceName=srv_resp["Service"]["Name"],
        HealthStatus="ALL",
    )
    assert len(instances["Instances"]) == 4
    assert set(inst["InstanceId"] for inst in instances["Instances"]) == {
        "i-123",
        "i-456",
        "i-789",
        "i-012",
    }

    instances = client.discover_instances(
        NamespaceName=ns_resp["Namespace"]["Name"],
        ServiceName=srv_resp["Service"]["Name"],
        HealthStatus="UNHEALTHY",
    )
    assert len(instances["Instances"]) == 2
    assert set(inst["InstanceId"] for inst in instances["Instances"]) == {
        "i-789",
        "i-012",
    }

    instances = client.discover_instances(
        NamespaceName=ns_resp["Namespace"]["Name"],
        ServiceName=srv_resp["Service"]["Name"],
        HealthStatus="HEALTHY",
    )
    assert len(instances["Instances"]) == 2
    assert set(inst["InstanceId"] for inst in instances["Instances"]) == {
        "i-123",
        "i-456",
    }

    instances = client.discover_instances(
        NamespaceName=ns_resp["Namespace"]["Name"],
        ServiceName=srv_resp["Service"]["Name"],
        HealthStatus="HEALTHY_OR_ELSE_ALL",
    )
    assert len(instances["Instances"]) == 2
    assert set(inst["InstanceId"] for inst in instances["Instances"]) == {
        "i-123",
        "i-456",
    }

    client.update_instance_custom_health_status(
        ServiceId=srv_resp["Service"]["Id"],
        InstanceId="i-123",
        Status="UNHEALTHY",
    )
    client.update_instance_custom_health_status(
        ServiceId=srv_resp["Service"]["Id"],
        InstanceId="i-456",
        Status="UNHEALTHY",
    )
    instances = client.discover_instances(
        NamespaceName=ns_resp["Namespace"]["Name"],
        ServiceName=srv_resp["Service"]["Name"],
        HealthStatus="HEALTHY_OR_ELSE_ALL",
    )
    assert len(instances["Instances"]) == 4
    assert set(inst["InstanceId"] for inst in instances["Instances"]) == {
        "i-123",
        "i-456",
        "i-789",
        "i-012",
    }


@mock_aws
def test_max_results_discover_instances(client, ns_resp, srv_resp):
    if not settings.TEST_DECORATOR_MODE:
        raise SkipTest(
            "Endpoint for discovering instances is prefixed with 'data-', and we can't intercept calls to 'data-localhost'"
        )

    instance_ids = ["i-123", "i-456", "i-789", "i-012"]
    for instance_id in instance_ids:
        client.register_instance(
            ServiceId=srv_resp["Service"]["Id"], InstanceId=instance_id, Attributes={}
        )

    instances = client.discover_instances(
        NamespaceName=ns_resp["Namespace"]["Name"],
        ServiceName=srv_resp["Service"]["Name"],
        MaxResults=2,
    )
    assert len(instances["Instances"]) == 2
    assert set(inst["InstanceId"] for inst in instances["Instances"]) == set(
        instance_ids[:2]
    )


@mock_aws
def test_discover_instances_revision(client, ns_resp, srv_resp):
    if not settings.TEST_DECORATOR_MODE:
        raise SkipTest(
            "Endpoint for discovering instances is prefixed with 'data-', and we can't intercept calls to 'data-localhost'"
        )

    instance_ids = ["i-123", "i-456", "i-789", "i-012"]
    for instance_id in instance_ids:
        client.register_instance(
            ServiceId=srv_resp["Service"]["Id"], InstanceId=instance_id, Attributes={}
        )

    revisions = client.discover_instances_revision(
        NamespaceName=ns_resp["Namespace"]["Name"],
        ServiceName=srv_resp["Service"]["Name"],
    )
    assert revisions["InstancesRevision"] == 4

    client.deregister_instance(ServiceId=srv_resp["Service"]["Id"], InstanceId="i-123")
    client.register_instance(
        ServiceId=srv_resp["Service"]["Id"], InstanceId="i-123", Attributes={}
    )
    revisions = client.discover_instances_revision(
        NamespaceName=ns_resp["Namespace"]["Name"],
        ServiceName=srv_resp["Service"]["Name"],
    )
    assert revisions["InstancesRevision"] == 6



from unittest import SkipTest

import boto3
import pytest
import requests
from responses import Response

from moto import mock_aws, settings
from moto.core.models import responses_mock


@mock_aws
@pytest.mark.network
def test_passthrough_requests() -> None:
    conn = boto3.client("sqs", region_name="us-west-1")
    conn.create_queue(QueueName="queue1")

    res = requests.get("https://google.com/")
    assert res.status_code < 400


if not settings.TEST_SERVER_MODE:

    @mock_aws
    def test_requests_to_amazon_subdomains_dont_work() -> None:
        res = requests.get("https://fakeservice.amazonaws.com/foo/bar")
        assert res.content == b"The method is not implemented"
        assert res.status_code == 400


@mock_aws
def test_decorator_ordering() -> None:
    """
    https://github.com/getmoto/moto/issues/3790#issuecomment-803979809
    """
    bucket_name = "banana-slugs"
    key = "trash-file"
    region = "us-east-1"
    client = boto3.client("s3", region_name=region)
    s3 = boto3.resource("s3", region_name=region)
    bucket = s3.Bucket(bucket_name)
    bucket.create()
    bucket.put_object(Body=b"ABCD", Key=key)
    presigned_url = client.generate_presigned_url(
        ClientMethod=client.get_object.__name__,
        Params={
            "Bucket": bucket_name,
            "Key": key,
            "ResponseContentDisposition": "attachment;filename=bar",
        },
    )

    resp = requests.get(presigned_url)
    assert resp.status_code == 200


@mock_aws()
def test_replace_and_remove_mock() -> None:
    if not settings.TEST_DECORATOR_MODE:
        raise SkipTest("Only need to test responses mock in decorator mode")
    rsp1 = Response(method="GET", url="http://example.com", body="test")
    responses_mock.add(rsp1)

    assert requests.get("http://example.com/").text == "test"

    rsp2 = Response(method="GET", url="http://example.com", body="test2")
    responses_mock.replace(rsp2)

    assert requests.get("http://example.com/").text == "test2"

    responses_mock.remove("GET", "https://example.com/")



import base64
import json
import uuid
from datetime import datetime, timedelta
from typing import Any, Dict, List, Optional, Union

from dateutil.tz import tzutc

from moto.core.base_backend import BackendDict, BaseBackend
from moto.core.common_models import BaseModel
from moto.moto_api._internal.managed_state_model import ManagedState
from moto.panorama.utils import (
    arn_formatter,
    deep_convert_datetime_to_isoformat,
    generate_package_id,
    hash_name,
)
from moto.utilities.paginator import paginate

from .exceptions import (
    ValidationError,
)

PAGINATION_MODEL = {
    "list_devices": {
        "input_token": "next_token",
        "limit_key": "max_results",
        "limit_default": 123,
        "unique_attribute": "device_id",
    },
    "list_nodes": {
        "input_token": "next_token",
        "limit_key": "max_results",
        "limit_default": 123,
        "unique_attribute": "package_id",
    },
    "list_application_instances": {
        "input_token": "next_token",
        "limit_key": "max_results",
        "limit_default": 123,
        "unique_attribute": "application_instance_id",
    },
}


class BaseObject(BaseModel):
    def camelCase(self, key: str) -> str:
        words = []
        for word in key.split("_"):
            words.append(word.title())
        return "".join(words)

    def update(self, details_json: str) -> None:
        details = json.loads(details_json)
        for k in details.keys():
            setattr(self, k, details[k])

    def gen_response_object(self) -> Dict[str, Any]:
        response_object: Dict[str, Any] = dict()
        for key, value in self.__dict__.items():
            if "_" in key:
                response_object[self.camelCase(key)] = value
            else:
                response_object[key[0].upper() + key[1:]] = value
        return response_object

    def response_object(self) -> Dict[str, Any]:
        return self.gen_response_object()


class Device(BaseObject):
    def __init__(
        self,
        account_id: str,
        region_name: str,
        description: Optional[str],
        name: str,
        network_configuration: Optional[Dict[str, Any]],
        tags: Optional[Dict[str, str]],
    ) -> None:
        # ManagedState is a class that helps us manage the state of a resource.
        # A Panorama Device has a lot of different states that has their own lifecycle.
        # To make all this states in the same object, and avoid changing ManagedState,
        # we use a ManagedState for each state and we manage them in the Device class.
        # Each ManagedState has a name composed with attribute name and device name to make subscription easier.
        self.__device_aggregated_status_manager = ManagedState(
            model_name=f"panorama::device_{name}_aggregated_status",
            transitions=[
                ("NOT-A-STATUS", "AWAITING_PROVISIONING"),
                ("AWAITING_PROVISIONING", "PENDING"),
                ("PENDING", "ONLINE"),
            ],
        )
        self.__device_provisioning_status_manager = ManagedState(
            model_name=f"panorama::device_{name}_provisioning_status",
            transitions=[
                ("NOT-A-STATUS", "AWAITING_PROVISIONING"),
                ("AWAITING_PROVISIONING", "PENDING"),
                ("PENDING", "SUCCEEDED"),
            ],
        )
        self.account_id = account_id
        self.region_name = region_name
        self.description = description
        self.name = name
        self.network_configuration = network_configuration
        self.tags = tags

        self.certificates = base64.b64encode("certificate".encode("utf-8")).decode(
            "utf-8"
        )
        self.arn = arn_formatter("device", self.name, self.account_id, self.region_name)
        self.device_id = f"device-{hash_name(name)}"
        self.iot_thing_name = ""

        self.alternate_softwares = [
            {"Version": "0.2.1"},
        ]
        self.brand: str = "AWS_PANORAMA"  # AWS_PANORAMA | LENOVO
        self.created_time = datetime.now(tzutc())
        self.last_updated_time = datetime.now(tzutc())
        self.current_networking_status = {
            "Ethernet0Status": {
                "ConnectionStatus": "CONNECTED",
                "HwAddress": "8C:0F:5F:60:F5:C4",
                "IpAddress": "192.168.1.300/24",
            },
            "Ethernet1Status": {
                "ConnectionStatus": "NOT_CONNECTED",
                "HwAddress": "8C:0F:6F:60:F4:F1",
                "IpAddress": "--",
            },
            "LastUpdatedTime": datetime.now(tzutc()),
            "NtpStatus": {
                "ConnectionStatus": "CONNECTED",
                "IpAddress": "91.224.149.41:123",
                "NtpServerName": "0.pool.ntp.org",
            },
        }
        self.current_software = "6.2.1"
        self.device_connection_status: str = "ONLINE"  # "ONLINE"|"OFFLINE"|"AWAITING_CREDENTIALS"|"NOT_AVAILABLE"|"ERROR"
        self.latest_device_job = {"JobType": "REBOOT", "Status": "COMPLETED"}
        self.latest_software = "6.2.1"
        self.lease_expiration_time = datetime.now(tzutc()) + timedelta(days=5)
        self.serial_number = "GAD81E29013274749"
        self.type: str = "PANORAMA_APPLIANCE"  # "PANORAMA_APPLIANCE_DEVELOPER_KIT", "PANORAMA_APPLIANCE"

    @property
    def device_aggregated_status(self) -> str:
        self.__device_aggregated_status_manager.advance()
        return self.__device_aggregated_status_manager.status  # type: ignore[return-value]

    @property
    def provisioning_status(self) -> str:
        self.__device_provisioning_status_manager.advance()
        return self.__device_provisioning_status_manager.status  # type: ignore[return-value]

    def response_object(self) -> Dict[str, Any]:
        response_object = super().gen_response_object()
        response_object = deep_convert_datetime_to_isoformat(response_object)
        static_response_fields = [
            "AlternateSoftwares",
            "Arn",
            "Brand",
            "CreatedTime",
            "CurrentNetworkingStatus",
            "CurrentSoftware",
            "Description",
            "DeviceConnectionStatus",
            "DeviceId",
            "LatestAlternateSoftware",
            "LatestDeviceJob",
            "LatestSoftware",
            "LeaseExpirationTime",
            "Name",
            "NetworkConfiguration",
            "SerialNumber",
            "Tags",
            "Type",
        ]
        return {
            **{
                k: v
                for k, v in response_object.items()
                if v is not None and k in static_response_fields
            },
            **{
                "DeviceAggregatedStatus": self.device_aggregated_status,
                "ProvisioningStatus": self.provisioning_status,
            },
        }

    def response_listed(self) -> Dict[str, Any]:
        response_object = super().gen_response_object()
        response_object = deep_convert_datetime_to_isoformat(response_object)
        static_response_fields = [
            "Brand",
            "CreatedTime",
            "CurrentSoftware",
            "Description",
            "DeviceId",
            "LastUpdatedTime",
            "LatestDeviceJob",
            "LeaseExpirationTime",
            "Name",
            "Tags",
            "Type",
        ]
        return {
            **{
                k: v
                for k, v in response_object.items()
                if v is not None and k in static_response_fields
            },
            **{
                "DeviceAggregatedStatus": self.device_aggregated_status,
                "ProvisioningStatus": self.provisioning_status,
            },
        }

    @property
    def response_provision(self) -> Dict[str, Union[str, bytes]]:
        return {
            "Arn": self.arn,
            "Certificates": self.certificates,
            "DeviceId": self.device_id,
            "IotThingName": self.iot_thing_name,
            "Status": self.provisioning_status,
        }

    @property
    def response_updated(self) -> Dict[str, str]:
        return {"DeviceId": self.device_id}

    @property
    def response_deleted(self) -> Dict[str, str]:
        return {"DeviceId": self.device_id}


class Package(BaseObject):
    def __init__(
        self,
        category: str,
        description: str,
        name: str,
        account_id: str,
        region_name: str,
        package_name: str,
        package_version: str,
    ):
        self.category = category
        self.description = description
        self.name = name
        now = datetime.now(tzutc())
        self.created_time = now
        self.last_updated_time = now
        self.package_name = package_name
        self.patch_version = generate_package_id(self.package_name)
        self.package_version = package_version
        self.output_package_name = f"{self.package_name}-{self.package_version}-{self.patch_version[:8]}-{self.name}"
        self.owner_account = account_id
        self.package_id = f"package-{hash_name(package_name)}"
        self.package_arn = arn_formatter(
            "package", self.package_id, account_id, region_name
        )

    def response_object(self) -> Dict[str, Any]:
        response_object = super().gen_response_object()
        response_object = deep_convert_datetime_to_isoformat(response_object)
        return response_object

    def response_listed(self) -> Dict[str, Any]:
        package_response = self.response_object()
        return package_response


class ApplicationInstance(BaseObject):
    def __init__(
        self,
        account_id: str,
        region_name: str,
        default_runtime_context_device: str,
        default_runtime_context_device_name: str,
        description: str,
        manifest_overrides_payload: Dict[str, str],
        manifest_payload: Dict[str, str],
        name: str,
        runtime_role_arn: str,
        tags: Dict[str, str],
    ) -> None:
        self.default_runtime_context_device = default_runtime_context_device
        self.default_runtime_context_device_name = default_runtime_context_device_name
        self.description = description
        self.manifest_overrides_payload = manifest_overrides_payload
        self.manifest_payload = manifest_payload
        self.name = name
        self.runtime_role_arn = runtime_role_arn
        self.tags = tags
        now = datetime.now(tzutc())
        self.created_time = now
        self.last_updated_time = now
        name = f"{self.name}-{self.created_time}"
        self.application_instance_id = f"applicationInstance-{hash_name(name).lower()}"
        self.arn = arn_formatter(
            "application-instance",
            self.application_instance_id,
            account_id=account_id,
            region_name=region_name,
        )
        self.health_status = "RUNNING"
        self.status = "DEPLOYMENT_SUCCEEDED"
        self.status_description = "string"
        self.runtime_context_states = [
            {
                "DesiredState": "RUNNING",
                "DeviceReportedStatus": "RUNNING",
                "DeviceReportedTime": now,
                "RuntimeContextName": "string",
            },
        ]

    def add_new_runtime_context_states(
        self, desired_state: str, device_reported_status: str
    ) -> None:
        now = datetime.now(tzutc())
        self.runtime_context_states.append(
            {
                "DesiredState": desired_state,
                "DeviceReportedStatus": device_reported_status,
                "DeviceReportedTime": now,
                "RuntimeContextName": "string",
            }
        )

    def response_object(self) -> Dict[str, Any]:
        response_object = super().gen_response_object()
        response_object = deep_convert_datetime_to_isoformat(response_object)
        return response_object

    def response_listed(self) -> Dict[str, Any]:
        package_response = self.response_object()
        return package_response

    def response_created(self) -> Dict[str, str]:
        return {"ApplicationInstanceId": self.application_instance_id}

    def response_describe(self) -> Dict[str, str]:
        response_object = self.response_object()
        return response_object


class Node(BaseObject):
    def __init__(
        self,
        job_id: str,
        job_tags: List[Dict[str, Union[str, Dict[str, str]]]],
        node_description: str,
        node_name: str,
        output_package_name: str,
        output_package_version: str,
        template_parameters: Dict[str, str],
        template_type: str,
    ) -> None:
        self.job_id = job_id
        now = datetime.now(tzutc())
        self.created_time = now
        self.last_updated_time = now
        self.job_tags = job_tags
        self.node_description = node_name
        self.status = "PENDING"
        self.node_name = node_name
        self.output_package_name = output_package_name
        self.output_package_version = output_package_version
        self.template_parameters = self.protect_secrets(template_parameters)
        self.template_type = template_type

    def response_object(self) -> Dict[str, Any]:
        response_object = super().gen_response_object()
        response_object = deep_convert_datetime_to_isoformat(response_object)
        return response_object

    def response_created(self) -> Dict[str, str]:
        return {"JobId": self.job_id}

    def response_described(self) -> Dict[str, Any]:
        return self.response_object()

    @staticmethod
    def protect_secrets(template_parameters: Dict[str, str]) -> Dict[str, str]:
        for key in template_parameters.keys():
            if key.lower() == "password":
                template_parameters[key] = "SAVED_AS_SECRET"
            if key.lower() == "username":
                template_parameters[key] = "SAVED_AS_SECRET"
        return template_parameters


class PanoramaBackend(BaseBackend):
    def __init__(self, region_name: str, account_id: str):
        super().__init__(region_name, account_id)
        self.devices_memory: Dict[str, Device] = {}
        self.node_from_template_memory: Dict[str, Node] = {}
        self.nodes_memory: Dict[str, Package] = {}
        self.application_instances_memory: Dict[str, ApplicationInstance] = {}

    def provision_device(
        self,
        description: Optional[str],
        name: str,
        networking_configuration: Optional[Dict[str, Any]],
        tags: Optional[Dict[str, str]],
    ) -> Device:
        device_obj = Device(
            account_id=self.account_id,
            region_name=self.region_name,
            description=description,
            name=name,
            network_configuration=networking_configuration,
            tags=tags,
        )

        self.devices_memory[device_obj.device_id] = device_obj
        return device_obj

    def describe_device(self, device_id: str) -> Device:
        device = self.devices_memory.get(device_id)
        if device is None:
            raise ValidationError(f"Device {device_id} not found")
        return device

    @paginate(pagination_model=PAGINATION_MODEL)
    def list_devices(
        self,
        device_aggregated_status_filter: str,
        name_filter: str,
        sort_by: str,  # "DEVICE_ID", "CREATED_TIME", "NAME", "DEVICE_AGGREGATED_STATUS"
        sort_order: str,  # "ASCENDING", "DESCENDING"
    ) -> List[Device]:
        devices_list = list(
            filter(
                lambda x: (name_filter is None or x.name.startswith(name_filter))
                and (
                    device_aggregated_status_filter is None
                    or x.device_aggregated_status == device_aggregated_status_filter
                ),
                self.devices_memory.values(),
            )
        )
        devices_list = list(
            sorted(
                devices_list,
                key={
                    "DEVICE_ID": lambda x: x.device_id,
                    "CREATED_TIME": lambda x: x.created_time,
                    "NAME": lambda x: x.name,
                    "DEVICE_AGGREGATED_STATUS": lambda x: x.device_aggregated_status,
                    None: lambda x: x.created_time,
                }[sort_by],
                reverse=sort_order == "DESCENDING",
            )
        )
        return devices_list

    def update_device_metadata(self, device_id: str, description: str) -> Device:
        self.devices_memory[device_id].description = description
        return self.devices_memory[device_id]

    def delete_device(self, device_id: str) -> Device:
        return self.devices_memory.pop(device_id)

    def create_node_from_template_job(
        self,
        job_tags: List[Dict[str, Union[str, Dict[str, str]]]],
        node_description: str,
        node_name: str,
        output_package_name: str,
        output_package_version: str,
        template_parameters: Dict[str, str],
        template_type: str,
    ) -> Node:
        job_id = str(uuid.uuid4()).lower()
        self.node_from_template_memory[job_id] = Node(
            job_id=job_id,
            job_tags=job_tags,
            node_description=node_description,
            node_name=node_name,
            output_package_name=output_package_name,
            output_package_version=output_package_version,
            template_parameters=template_parameters,
            template_type=template_type,
        )
        if template_type == "RTSP_CAMERA_STREAM":
            package = Package(
                category="MEDIA_SOURCE",
                description=node_description,
                name=node_name,
                account_id=self.account_id,
                region_name=self.region_name,
                package_name=output_package_name,
                package_version=output_package_version,
            )
            self.nodes_memory[package.package_id] = package

        return self.node_from_template_memory[job_id]

    def describe_node_from_template_job(self, job_id: str) -> Node:
        return self.node_from_template_memory[job_id]

    @paginate(pagination_model=PAGINATION_MODEL)
    def list_nodes(self, category: str) -> List[Package]:
        category_nodes = list(
            filter(
                lambda x: x.category == category,
                self.nodes_memory.values(),
            )
        )
        return category_nodes

    def create_application_instance(
        self,
        application_instance_id_to_replace: Optional[str],
        default_runtime_context_device: str,
        description: str,
        manifest_overrides_payload: Dict[str, str],
        manifest_payload: Dict[str, str],
        name: str,
        runtime_role_arn: str,
        tags: Dict[str, str],
    ) -> ApplicationInstance:
        device = self.devices_memory.get(default_runtime_context_device)
        if device is None:
            raise ValidationError(f"Device {default_runtime_context_device} not found")
        if (
            application_instance_id_to_replace
            and application_instance_id_to_replace in self.application_instances_memory
        ):
            removed_application_instance = self.application_instances_memory[
                application_instance_id_to_replace
            ]
            removed_application_instance.status = "REMOVAL_SUCCEEDED"
            removed_application_instance.add_new_runtime_context_states(
                desired_state="REMOVED", device_reported_status="REMOVAL_IN_PROGRESS"
            )

        application_instance = ApplicationInstance(
            account_id=self.account_id,
            region_name=self.region_name,
            default_runtime_context_device=default_runtime_context_device,
            default_runtime_context_device_name=device.name,
            description=description,
            manifest_overrides_payload=manifest_overrides_payload,
            manifest_payload=manifest_payload,
            name=name,
            runtime_role_arn=runtime_role_arn,
            tags=tags,
        )
        self.application_instances_memory[
            application_instance.application_instance_id
        ] = application_instance
        return application_instance

    def describe_application_instance(
        self, application_instance_id: str
    ) -> ApplicationInstance:
        application_instance = self.application_instances_memory[
            application_instance_id
        ]
        return application_instance

    @paginate(pagination_model=PAGINATION_MODEL)
    def list_application_instances(
        self,
        device_id: Optional[str],
        status_filter: Optional[str],
    ) -> List[ApplicationInstance]:
        filtered_application_instances = filter(
            lambda x: x.status == status_filter if status_filter else True,
            filter(
                lambda x: x.default_runtime_context_device == device_id
                if device_id
                else True,
                self.application_instances_memory.values(),
            ),
        )
        return list(filtered_application_instances)


panorama_backends = BackendDict(
    PanoramaBackend,
    "panorama",
    False,
    additional_regions=[
        "us-east-1",
        "us-west-2",
        "ca-central-1",
        "eu-west-1",
        "ap-southeast-2",
        "ap-southeast-1",
    ],
)



import gzip
import json
from time import sleep
from uuid import uuid4

import boto3
import pytest
from botocore.exceptions import ClientError

from . import dynamodb_aws_verified


@pytest.mark.aws_verified
@dynamodb_aws_verified(create_table=False)
def test_import_from_missing_s3_table(table_name=None):
    client = boto3.client("dynamodb", region_name="us-east-1")

    table_name = "t" + str(uuid4())[0:6]

    import_description = client.import_table(
        S3BucketSource={"S3Bucket": f"{uuid4()}"},
        InputFormat="DYNAMODB_JSON",
        InputCompressionType="NONE",
        TableCreationParameters={
            "TableName": table_name,
            "AttributeDefinitions": [
                {"AttributeName": "pk", "AttributeType": "S"},
            ],
            "KeySchema": [
                {"AttributeName": "pk", "KeyType": "HASH"},
            ],
            "BillingMode": "PAY_PER_REQUEST",
        },
    )["ImportTableDescription"]

    import_details = wait_for_import(client, import_description)

    assert import_details["ImportStatus"] == "FAILED"
    assert import_details["TableArn"].endswith(f":table/{table_name}")
    assert import_details["FailureCode"] == "S3NoSuchBucket"
    assert "The specified bucket does not exist" in import_details["FailureMessage"]

    assert table_name not in client.list_tables()["TableNames"]


@pytest.mark.aws_verified
@dynamodb_aws_verified(create_table=False)
def test_import_has_regular_validation(table_name=None):
    client = boto3.client("dynamodb", region_name="us-east-1")

    table_name = "t" + str(uuid4())[0:6]

    with pytest.raises(ClientError) as exc:
        client.import_table(
            S3BucketSource={"S3Bucket": f"{uuid4()}"},
            InputFormat="DYNAMODB_JSON",
            InputCompressionType="NONE",
            TableCreationParameters={
                "TableName": table_name,
                "AttributeDefinitions": [
                    {"AttributeName": "pk", "AttributeType": "S"},
                ],
                "KeySchema": [
                    {"AttributeName": "pk", "KeyType": "HASH"},
                ],
                "BillingMode": "PROVISIONED",
            },
        )

    err = exc.value.response["Error"]
    assert (
        err["Message"]
        == "One or more parameter values were invalid: ReadCapacityUnits and WriteCapacityUnits must both be specified when BillingMode is PROVISIONED"
    )

    assert table_name not in client.list_tables()["TableNames"]


@pytest.mark.aws_verified
@dynamodb_aws_verified(create_table=False)
def test_import_from_empty_s3_bucket(table_name=None):
    client = boto3.client("dynamodb", region_name="us-east-1")
    s3 = boto3.client("s3", region_name="us-east-1")

    s3_bucket_name = f"inttest{uuid4()}"
    table_name = "moto_test_" + str(uuid4())[0:6]

    s3.create_bucket(Bucket=s3_bucket_name)

    import_description = client.import_table(
        S3BucketSource={"S3Bucket": s3_bucket_name},
        InputFormat="DYNAMODB_JSON",
        InputCompressionType="NONE",
        TableCreationParameters={
            "TableName": table_name,
            "AttributeDefinitions": [
                {"AttributeName": "pk", "AttributeType": "S"},
            ],
            "KeySchema": [
                {"AttributeName": "pk", "KeyType": "HASH"},
            ],
            "BillingMode": "PAY_PER_REQUEST",
        },
    )["ImportTableDescription"]

    import_details = wait_for_import(client, import_description)

    assert import_details["ImportStatus"] == "COMPLETED"
    assert import_details["ErrorCount"] == 0
    assert import_details["ProcessedItemCount"] == 0
    assert import_details["ImportedItemCount"] == 0
    assert import_details["ProcessedSizeBytes"] == 0

    assert table_name in client.list_tables()["TableNames"]

    assert client.scan(TableName=table_name)["Items"] == []

    client.delete_table(TableName=table_name)

    s3.delete_bucket(Bucket=s3_bucket_name)


@pytest.mark.aws_verified
@dynamodb_aws_verified(create_table=False)
def test_import_table_single_file_with_multiple_items():
    client = boto3.client("dynamodb", region_name="us-east-2")
    s3 = boto3.client("s3", region_name="us-east-1")

    s3_bucket_name = f"inttest{uuid4()}"
    table_name = "moto_test_" + str(uuid4())[0:6]

    s3.create_bucket(Bucket=s3_bucket_name)

    data = ""
    for i in range(5):
        data += (
            json.dumps({"Item": {"pk": {"S": f"msg{i}"}, "data": {"S": f"{uuid4()}"}}})
            + "\n"
        )
    for i in range(10, 15):
        data += json.dumps(
            {"Item": {"pk": {"S": f"msg{i}"}, "data": {"S": f"{uuid4()}"}}}
        )
    filename1 = "data.json"
    s3.put_object(
        Bucket=s3_bucket_name,
        Body=data,
        Key=filename1,
    )

    import_description = client.import_table(
        S3BucketSource={"S3Bucket": s3_bucket_name},
        InputFormat="DYNAMODB_JSON",
        InputCompressionType="NONE",
        TableCreationParameters={
            "TableName": table_name,
            "AttributeDefinitions": [
                {"AttributeName": "pk", "AttributeType": "S"},
            ],
            "KeySchema": [
                {"AttributeName": "pk", "KeyType": "HASH"},
            ],
            "BillingMode": "PAY_PER_REQUEST",
        },
    )["ImportTableDescription"]

    import_details = wait_for_import(client, import_description)

    assert import_details["ImportStatus"] == "COMPLETED"
    assert import_details["ErrorCount"] == 0
    assert import_details["ProcessedItemCount"] == 10
    assert import_details["ImportedItemCount"] == 10
    assert import_details["ProcessedSizeBytes"] > 0

    assert table_name in client.list_tables()["TableNames"]
    expected = [
        "msg0",
        "msg1",
        "msg10",
        "msg11",
        "msg12",
        "msg13",
        "msg14",
        "msg2",
        "msg3",
        "msg4",
    ]
    assert (
        sorted([i["pk"]["S"] for i in client.scan(TableName=table_name)["Items"]])
        == expected
    )

    client.delete_table(TableName=table_name)

    s3.delete_object(Bucket=s3_bucket_name, Key=filename1)
    s3.delete_bucket(Bucket=s3_bucket_name)


@pytest.mark.aws_verified
@dynamodb_aws_verified(create_table=False)
def test_import_table_multiple_files():
    client = boto3.client("dynamodb", region_name="us-east-2")
    s3 = boto3.client("s3", region_name="us-east-1")

    s3_bucket_name = f"inttest{uuid4()}"
    table_name = "moto_test_" + str(uuid4())[0:6]

    s3.create_bucket(Bucket=s3_bucket_name)

    items_file1 = {"Item": {"pk": {"S": "msg1"}, "data": {"S": f"{uuid4()}"}}}
    filename1 = "data.json"
    s3.put_object(
        Bucket=s3_bucket_name,
        Body=json.dumps(items_file1),
        Key=filename1,
    )

    items_file2 = {"Item": {"pk": {"S": "msg2"}, "data": {"S": f"{uuid4()}"}}}
    filename2 = "completely_random_filename_without_extension"
    s3.put_object(
        Bucket=s3_bucket_name,
        Body=json.dumps(items_file2),
        Key=filename2,
    )

    import_description = client.import_table(
        S3BucketSource={"S3Bucket": s3_bucket_name},
        InputFormat="DYNAMODB_JSON",
        InputCompressionType="NONE",
        TableCreationParameters={
            "TableName": table_name,
            "AttributeDefinitions": [
                {"AttributeName": "pk", "AttributeType": "S"},
            ],
            "KeySchema": [
                {"AttributeName": "pk", "KeyType": "HASH"},
            ],
            "BillingMode": "PAY_PER_REQUEST",
        },
    )["ImportTableDescription"]

    import_details = wait_for_import(client, import_description)

    assert import_details["ImportStatus"] == "COMPLETED"
    assert import_details["ErrorCount"] == 0
    assert import_details["ProcessedItemCount"] == 2
    assert import_details["ImportedItemCount"] == 2
    assert import_details["ProcessedSizeBytes"] > 0

    assert table_name in client.list_tables()["TableNames"]

    assert [i["pk"]["S"] for i in client.scan(TableName=table_name)["Items"]] == [
        "msg1",
        "msg2",
    ]

    client.delete_table(TableName=table_name)

    s3.delete_object(Bucket=s3_bucket_name, Key=filename1)
    s3.delete_object(Bucket=s3_bucket_name, Key=filename2)
    s3.delete_bucket(Bucket=s3_bucket_name)


@pytest.mark.aws_verified
@dynamodb_aws_verified(create_table=False)
def test_some_successfull_files_and_some_with_unknown_data():
    client = boto3.client("dynamodb", region_name="us-east-2")
    s3 = boto3.client("s3", region_name="us-east-1")

    s3_bucket_name = f"inttest{uuid4()}"
    table_name = "moto_test_" + str(uuid4())[0:6]

    s3.create_bucket(Bucket=s3_bucket_name)

    items_file1 = {"Item": {"pk": {"S": "msg1"}, "data": {"S": f"{uuid4()}"}}}
    filename1 = "data.json"
    s3.put_object(
        Bucket=s3_bucket_name,
        Body=json.dumps(items_file1),
        Key=filename1,
    )

    items_file2 = {"pk": {"S": "msg2"}, "data": {"S": f"{uuid4()}"}}
    filename2 = "invaliddata"
    s3.put_object(
        Bucket=s3_bucket_name,
        Body=json.dumps(items_file2),
        Key=filename2,
    )

    import_description = client.import_table(
        S3BucketSource={"S3Bucket": s3_bucket_name},
        InputFormat="DYNAMODB_JSON",
        InputCompressionType="NONE",
        TableCreationParameters={
            "TableName": table_name,
            "AttributeDefinitions": [
                {"AttributeName": "pk", "AttributeType": "S"},
            ],
            "KeySchema": [
                {"AttributeName": "pk", "KeyType": "HASH"},
            ],
            "BillingMode": "PAY_PER_REQUEST",
        },
    )["ImportTableDescription"]

    import_details = wait_for_import(client, import_description)

    assert import_details["ImportStatus"] == "FAILED"
    assert import_details["ErrorCount"] == 1
    assert import_details["ProcessedItemCount"] == 2
    assert import_details["ImportedItemCount"] == 1
    assert import_details["ProcessedSizeBytes"] > 0

    assert table_name in client.list_tables()["TableNames"]

    assert [i["pk"]["S"] for i in client.scan(TableName=table_name)["Items"]] == [
        "msg1"
    ]

    client.delete_table(TableName=table_name)

    s3.delete_object(Bucket=s3_bucket_name, Key=filename1)
    s3.delete_object(Bucket=s3_bucket_name, Key=filename2)
    s3.delete_bucket(Bucket=s3_bucket_name)


@pytest.mark.aws_verified
@dynamodb_aws_verified(create_table=False)
def test_only_process_file_with_prefix():
    client = boto3.client("dynamodb", region_name="us-east-2")
    s3 = boto3.client("s3", region_name="us-east-1")

    s3_bucket_name = f"inttest{uuid4()}"
    table_name = "moto_test_" + str(uuid4())[0:6]

    s3.create_bucket(Bucket=s3_bucket_name)

    items_file1 = {"Item": {"pk": {"S": "msg1"}, "data": {"S": f"{uuid4()}"}}}
    filename1 = "yesdata.json"
    s3.put_object(
        Bucket=s3_bucket_name,
        Body=json.dumps(items_file1),
        Key=filename1,
    )

    items_file2 = {"Item": {"pk": {"S": "msg2"}, "data": {"S": f"{uuid4()}"}}}
    filename2 = "nodata.json"
    s3.put_object(
        Bucket=s3_bucket_name,
        Body=json.dumps(items_file2),
        Key=filename2,
    )

    import_description = client.import_table(
        S3BucketSource={"S3Bucket": s3_bucket_name, "S3KeyPrefix": "yes"},
        InputFormat="DYNAMODB_JSON",
        InputCompressionType="NONE",
        TableCreationParameters={
            "TableName": table_name,
            "AttributeDefinitions": [
                {"AttributeName": "pk", "AttributeType": "S"},
            ],
            "KeySchema": [
                {"AttributeName": "pk", "KeyType": "HASH"},
            ],
            "BillingMode": "PAY_PER_REQUEST",
        },
    )["ImportTableDescription"]

    import_details = wait_for_import(client, import_description)

    assert import_details["ImportStatus"] == "COMPLETED"
    assert import_details["ErrorCount"] == 0
    assert import_details["ProcessedItemCount"] == 1
    assert import_details["ImportedItemCount"] == 1
    assert import_details["ProcessedSizeBytes"] > 0

    assert table_name in client.list_tables()["TableNames"]

    assert [i["pk"]["S"] for i in client.scan(TableName=table_name)["Items"]] == [
        "msg1"
    ]

    client.delete_table(TableName=table_name)

    s3.delete_object(Bucket=s3_bucket_name, Key=filename1)
    s3.delete_object(Bucket=s3_bucket_name, Key=filename2)
    s3.delete_bucket(Bucket=s3_bucket_name)


@pytest.mark.aws_verified
@dynamodb_aws_verified(create_table=False)
def test_process_gzipped_file():
    client = boto3.client("dynamodb", region_name="us-east-2")
    s3 = boto3.client("s3", region_name="us-east-1")

    s3_bucket_name = f"inttest{uuid4()}"
    table_name = "moto_test_" + str(uuid4())[0:6]

    s3.create_bucket(Bucket=s3_bucket_name)

    items_file1 = {"Item": {"pk": {"S": "msg1"}, "data": {"S": f"{uuid4()}"}}}
    filename1 = "data.json"
    s3.put_object(
        Bucket=s3_bucket_name,
        Body=gzip.compress(json.dumps(items_file1).encode("utf-8")),
        Key=filename1,
    )

    import_description = client.import_table(
        S3BucketSource={"S3Bucket": s3_bucket_name},
        InputFormat="DYNAMODB_JSON",
        InputCompressionType="GZIP",
        TableCreationParameters={
            "TableName": table_name,
            "AttributeDefinitions": [
                {"AttributeName": "pk", "AttributeType": "S"},
            ],
            "KeySchema": [
                {"AttributeName": "pk", "KeyType": "HASH"},
            ],
            "BillingMode": "PAY_PER_REQUEST",
        },
    )["ImportTableDescription"]

    import_details = wait_for_import(client, import_description)

    assert import_details["ImportStatus"] == "COMPLETED"
    assert import_details["ErrorCount"] == 0
    assert import_details["ProcessedItemCount"] == 1
    assert import_details["ImportedItemCount"] == 1
    assert import_details["ProcessedSizeBytes"] > 0

    assert table_name in client.list_tables()["TableNames"]

    assert [i["pk"]["S"] for i in client.scan(TableName=table_name)["Items"]] == [
        "msg1"
    ]

    client.delete_table(TableName=table_name)

    s3.delete_object(Bucket=s3_bucket_name, Key=filename1)
    s3.delete_bucket(Bucket=s3_bucket_name)


def wait_for_import(client, import_description):
    import_details = client.describe_import(ImportArn=import_description["ImportArn"])
    status = import_details["ImportTableDescription"]["ImportStatus"]
    while status == "IN_PROGRESS":
        sleep(0.1)
        import_details = client.describe_import(
            ImportArn=import_description["ImportArn"]
        )
        status = import_details["ImportTableDescription"]["ImportStatus"]
    return import_details["ImportTableDescription"]



import json

from moto.core.responses import BaseResponse

from .models import KinesisBackend, kinesis_backends


class KinesisResponse(BaseResponse):
    def __init__(self) -> None:
        super().__init__(service_name="kinesis")

    @property
    def kinesis_backend(self) -> KinesisBackend:
        return kinesis_backends[self.current_account][self.region]

    def create_stream(self) -> str:
        stream_name = self._get_param("StreamName")
        shard_count = self._get_param("ShardCount")
        stream_mode = self._get_param("StreamModeDetails")
        self.kinesis_backend.create_stream(
            stream_name, shard_count, stream_mode=stream_mode
        )
        return ""

    def describe_stream(self) -> str:
        stream_name = self._get_param("StreamName")
        stream_arn = self._get_param("StreamARN")
        limit = self._get_param("Limit")
        stream = self.kinesis_backend.describe_stream(stream_arn, stream_name)
        return json.dumps(stream.to_json(shard_limit=limit))

    def describe_stream_summary(self) -> str:
        stream_arn = self._get_param("StreamARN")
        stream_name = self._get_param("StreamName")
        stream = self.kinesis_backend.describe_stream_summary(stream_arn, stream_name)
        return json.dumps(stream.to_json_summary())

    def list_streams(self) -> str:
        streams = self.kinesis_backend.list_streams()
        stream_names = [stream.stream_name for stream in streams]
        max_streams = self._get_param("Limit", 10)
        try:
            token = self._get_param("ExclusiveStartStreamName")
        except ValueError:
            token = self._get_param("ExclusiveStartStreamName")
        if token:
            start = stream_names.index(token) + 1
        else:
            start = 0
        streams_resp = stream_names[start : start + max_streams]
        has_more_streams = False
        if start + max_streams < len(stream_names):
            has_more_streams = True

        return json.dumps(
            {
                "HasMoreStreams": has_more_streams,
                "StreamNames": streams_resp,
                "StreamSummaries": [
                    stream.to_json_summary()["StreamDescriptionSummary"]
                    for stream in streams
                ],
            }
        )

    def delete_stream(self) -> str:
        stream_arn = self._get_param("StreamARN")
        stream_name = self._get_param("StreamName")
        self.kinesis_backend.delete_stream(stream_arn, stream_name)
        return ""

    def get_shard_iterator(self) -> str:
        stream_arn = self._get_param("StreamARN")
        stream_name = self._get_param("StreamName")
        shard_id = self._get_param("ShardId")
        shard_iterator_type = self._get_param("ShardIteratorType")
        starting_sequence_number = self._get_param("StartingSequenceNumber")
        at_timestamp = self._get_param("Timestamp")

        shard_iterator = self.kinesis_backend.get_shard_iterator(
            stream_arn,
            stream_name,
            shard_id,
            shard_iterator_type,
            starting_sequence_number,
            at_timestamp,
        )

        return json.dumps({"ShardIterator": shard_iterator})

    def get_records(self) -> str:
        stream_arn = self._get_param("StreamARN")
        shard_iterator = self._get_param("ShardIterator")
        limit = self._get_param("Limit")

        (
            next_shard_iterator,
            records,
            millis_behind_latest,
        ) = self.kinesis_backend.get_records(stream_arn, shard_iterator, limit)

        return json.dumps(
            {
                "NextShardIterator": next_shard_iterator,
                "Records": [record.to_json() for record in records],
                "MillisBehindLatest": millis_behind_latest,
            }
        )

    def put_record(self) -> str:
        stream_arn = self._get_param("StreamARN")
        stream_name = self._get_param("StreamName")
        partition_key = self._get_param("PartitionKey")
        explicit_hash_key = self._get_param("ExplicitHashKey")
        data = self._get_param("Data")

        sequence_number, shard_id = self.kinesis_backend.put_record(
            stream_arn,
            stream_name,
            partition_key,
            explicit_hash_key,
            data,
        )

        return json.dumps({"SequenceNumber": sequence_number, "ShardId": shard_id})

    def put_records(self) -> str:
        stream_arn = self._get_param("StreamARN")
        stream_name = self._get_param("StreamName")
        records = self._get_param("Records")

        response = self.kinesis_backend.put_records(stream_arn, stream_name, records)

        return json.dumps(response)

    def split_shard(self) -> str:
        stream_arn = self._get_param("StreamARN")
        stream_name = self._get_param("StreamName")
        shard_to_split = self._get_param("ShardToSplit")
        new_starting_hash_key = self._get_param("NewStartingHashKey")
        self.kinesis_backend.split_shard(
            stream_arn, stream_name, shard_to_split, new_starting_hash_key
        )
        return ""

    def merge_shards(self) -> str:
        stream_arn = self._get_param("StreamARN")
        stream_name = self._get_param("StreamName")
        shard_to_merge = self._get_param("ShardToMerge")
        adjacent_shard_to_merge = self._get_param("AdjacentShardToMerge")
        self.kinesis_backend.merge_shards(
            stream_arn, stream_name, shard_to_merge, adjacent_shard_to_merge
        )
        return ""

    def list_shards(self) -> str:
        stream_arn = self._get_param("StreamARN")
        stream_name = self._get_param("StreamName")
        next_token = self._get_param("NextToken")
        max_results = self._get_param("MaxResults", 10000)
        shards, token = self.kinesis_backend.list_shards(
            stream_arn=stream_arn,
            stream_name=stream_name,
            limit=max_results,
            next_token=next_token,
        )
        res = {"Shards": [s.to_json() for s in shards], "NextToken": token}
        return json.dumps(res)

    def update_shard_count(self) -> str:
        stream_arn = self._get_param("StreamARN")
        stream_name = self._get_param("StreamName")
        target_shard_count = self._get_param("TargetShardCount")
        current_shard_count = self.kinesis_backend.update_shard_count(
            stream_arn=stream_arn,
            stream_name=stream_name,
            target_shard_count=target_shard_count,
        )
        return json.dumps(
            dict(
                StreamName=stream_name,
                CurrentShardCount=current_shard_count,
                TargetShardCount=target_shard_count,
            )
        )

    def increase_stream_retention_period(self) -> str:
        stream_arn = self._get_param("StreamARN")
        stream_name = self._get_param("StreamName")
        retention_period_hours = self._get_param("RetentionPeriodHours")
        self.kinesis_backend.increase_stream_retention_period(
            stream_arn, stream_name, retention_period_hours
        )
        return ""

    def decrease_stream_retention_period(self) -> str:
        stream_arn = self._get_param("StreamARN")
        stream_name = self._get_param("StreamName")
        retention_period_hours = self._get_param("RetentionPeriodHours")
        self.kinesis_backend.decrease_stream_retention_period(
            stream_arn, stream_name, retention_period_hours
        )
        return ""

    def add_tags_to_stream(self) -> str:
        stream_arn = self._get_param("StreamARN")
        stream_name = self._get_param("StreamName")
        tags = self._get_param("Tags")
        self.kinesis_backend.add_tags_to_stream(stream_arn, stream_name, tags)
        return json.dumps({})

    def list_tags_for_stream(self) -> str:
        stream_arn = self._get_param("StreamARN")
        stream_name = self._get_param("StreamName")
        exclusive_start_tag_key = self._get_param("ExclusiveStartTagKey")
        limit = self._get_param("Limit")
        response = self.kinesis_backend.list_tags_for_stream(
            stream_arn, stream_name, exclusive_start_tag_key, limit
        )
        return json.dumps(response)

    def remove_tags_from_stream(self) -> str:
        stream_arn = self._get_param("StreamARN")
        stream_name = self._get_param("StreamName")
        tag_keys = self._get_param("TagKeys")
        self.kinesis_backend.remove_tags_from_stream(stream_arn, stream_name, tag_keys)
        return json.dumps({})

    def enable_enhanced_monitoring(self) -> str:
        stream_arn = self._get_param("StreamARN")
        stream_name = self._get_param("StreamName")
        shard_level_metrics = self._get_param("ShardLevelMetrics")
        arn, name, current, desired = self.kinesis_backend.enable_enhanced_monitoring(
            stream_arn=stream_arn,
            stream_name=stream_name,
            shard_level_metrics=shard_level_metrics,
        )
        return json.dumps(
            dict(
                StreamName=name,
                CurrentShardLevelMetrics=current,
                DesiredShardLevelMetrics=desired,
                StreamARN=arn,
            )
        )

    def disable_enhanced_monitoring(self) -> str:
        stream_arn = self._get_param("StreamARN")
        stream_name = self._get_param("StreamName")
        shard_level_metrics = self._get_param("ShardLevelMetrics")
        arn, name, current, desired = self.kinesis_backend.disable_enhanced_monitoring(
            stream_arn=stream_arn,
            stream_name=stream_name,
            to_be_disabled=shard_level_metrics,
        )
        return json.dumps(
            dict(
                StreamName=name,
                CurrentShardLevelMetrics=current,
                DesiredShardLevelMetrics=desired,
                StreamARN=arn,
            )
        )

    def list_stream_consumers(self) -> str:
        stream_arn = self._get_param("StreamARN")
        consumers = self.kinesis_backend.list_stream_consumers(stream_arn=stream_arn)
        return json.dumps(dict(Consumers=[c.to_json() for c in consumers]))

    def register_stream_consumer(self) -> str:
        stream_arn = self._get_param("StreamARN")
        consumer_name = self._get_param("ConsumerName")
        consumer = self.kinesis_backend.register_stream_consumer(
            stream_arn=stream_arn, consumer_name=consumer_name
        )
        return json.dumps(dict(Consumer=consumer.to_json()))

    def describe_stream_consumer(self) -> str:
        stream_arn = self._get_param("StreamARN")
        consumer_name = self._get_param("ConsumerName")
        consumer_arn = self._get_param("ConsumerARN")
        consumer = self.kinesis_backend.describe_stream_consumer(
            stream_arn=stream_arn,
            consumer_name=consumer_name,
            consumer_arn=consumer_arn,
        )
        return json.dumps(
            dict(ConsumerDescription=consumer.to_json(include_stream_arn=True))
        )

    def deregister_stream_consumer(self) -> str:
        stream_arn = self._get_param("StreamARN")
        consumer_name = self._get_param("ConsumerName")
        consumer_arn = self._get_param("ConsumerARN")
        self.kinesis_backend.deregister_stream_consumer(
            stream_arn=stream_arn,
            consumer_name=consumer_name,
            consumer_arn=consumer_arn,
        )
        return json.dumps(dict())

    def start_stream_encryption(self) -> str:
        stream_arn = self._get_param("StreamARN")
        stream_name = self._get_param("StreamName")
        encryption_type = self._get_param("EncryptionType")
        key_id = self._get_param("KeyId")
        self.kinesis_backend.start_stream_encryption(
            stream_arn=stream_arn,
            stream_name=stream_name,
            encryption_type=encryption_type,
            key_id=key_id,
        )
        return json.dumps(dict())

    def stop_stream_encryption(self) -> str:
        stream_arn = self._get_param("StreamARN")
        stream_name = self._get_param("StreamName")
        self.kinesis_backend.stop_stream_encryption(
            stream_arn=stream_arn, stream_name=stream_name
        )
        return json.dumps(dict())

    def update_stream_mode(self) -> str:
        stream_arn = self._get_param("StreamARN")
        stream_mode = self._get_param("StreamModeDetails")
        self.kinesis_backend.update_stream_mode(stream_arn, stream_mode)
        return "{}"

    def put_resource_policy(self) -> str:
        resource_arn = self._get_param("ResourceARN")
        policy = self._get_param("Policy")
        self.kinesis_backend.put_resource_policy(
            resource_arn=resource_arn,
            policy_doc=policy,
        )
        return json.dumps(dict())

    def get_resource_policy(self) -> str:
        resource_arn = self._get_param("ResourceARN")
        policy = self.kinesis_backend.get_resource_policy(resource_arn=resource_arn)
        return json.dumps({"Policy": policy})

    def delete_resource_policy(self) -> str:
        resource_arn = self._get_param("ResourceARN")
        self.kinesis_backend.delete_resource_policy(
            resource_arn=resource_arn,
        )
        return json.dumps(dict())



from datetime import datetime

import boto3
import pytest
from botocore.exceptions import ClientError

from moto import mock_aws
from moto.s3.responses import DEFAULT_REGION_NAME


@mock_aws
def test_lifecycle_with_filters():
    client = boto3.client("s3", DEFAULT_REGION_NAME)
    client.create_bucket(Bucket="bucket")

    # Create a lifecycle rule with a Filter (no tags):
    lfc = {
        "Rules": [
            {
                "Expiration": {"Days": 7},
                "ID": "wholebucket",
                "Filter": {"Prefix": ""},
                "Status": "Enabled",
            }
        ]
    }
    client.put_bucket_lifecycle_configuration(
        Bucket="bucket", LifecycleConfiguration=lfc
    )
    result = client.get_bucket_lifecycle_configuration(Bucket="bucket")
    assert len(result["Rules"]) == 1
    assert result["Rules"][0]["Filter"]["Prefix"] == ""
    assert not result["Rules"][0]["Filter"].get("And")
    assert not result["Rules"][0]["Filter"].get("Tag")
    with pytest.raises(KeyError):
        assert result["Rules"][0]["Prefix"]

    # Without any prefixes and an empty filter (this is by default a prefix for the whole bucket):
    lfc = {
        "Rules": [
            {
                "Expiration": {"Days": 7},
                "ID": "wholebucket",
                "Filter": {},
                "Status": "Enabled",
            }
        ]
    }
    client.put_bucket_lifecycle_configuration(
        Bucket="bucket", LifecycleConfiguration=lfc
    )
    result = client.get_bucket_lifecycle_configuration(Bucket="bucket")
    assert len(result["Rules"]) == 1
    with pytest.raises(KeyError):
        assert result["Rules"][0]["Prefix"]

    # If we remove the filter -- and don't specify a Prefix, then this is bad:
    lfc["Rules"][0].pop("Filter")
    with pytest.raises(ClientError) as err:
        client.put_bucket_lifecycle_configuration(
            Bucket="bucket", LifecycleConfiguration=lfc
        )
    assert err.value.response["Error"]["Code"] == "MalformedXML"

    # With a tag:
    lfc["Rules"][0]["Filter"] = {"Tag": {"Key": "mytag", "Value": "mytagvalue"}}
    client.put_bucket_lifecycle_configuration(
        Bucket="bucket", LifecycleConfiguration=lfc
    )
    result = client.get_bucket_lifecycle_configuration(Bucket="bucket")
    assert len(result["Rules"]) == 1
    with pytest.raises(KeyError):
        assert result["Rules"][0]["Filter"]["Prefix"]
    assert not result["Rules"][0]["Filter"].get("And")
    assert result["Rules"][0]["Filter"]["Tag"]["Key"] == "mytag"
    assert result["Rules"][0]["Filter"]["Tag"]["Value"] == "mytagvalue"
    with pytest.raises(KeyError):
        assert result["Rules"][0]["Prefix"]

    # With And (single tag):
    lfc["Rules"][0]["Filter"] = {
        "And": {
            "Prefix": "some/prefix",
            "Tags": [{"Key": "mytag", "Value": "mytagvalue"}],
        }
    }
    client.put_bucket_lifecycle_configuration(
        Bucket="bucket", LifecycleConfiguration=lfc
    )
    result = client.get_bucket_lifecycle_configuration(Bucket="bucket")
    assert len(result["Rules"]) == 1
    assert not result["Rules"][0]["Filter"].get("Prefix")
    assert result["Rules"][0]["Filter"]["And"]["Prefix"] == "some/prefix"
    assert len(result["Rules"][0]["Filter"]["And"]["Tags"]) == 1
    assert result["Rules"][0]["Filter"]["And"]["Tags"][0]["Key"] == "mytag"
    assert result["Rules"][0]["Filter"]["And"]["Tags"][0]["Value"] == "mytagvalue"
    with pytest.raises(KeyError):
        assert result["Rules"][0]["Prefix"]

    # With multiple And tags:
    lfc["Rules"][0]["Filter"]["And"] = {
        "Prefix": "some/prefix",
        "Tags": [
            {"Key": "mytag", "Value": "mytagvalue"},
            {"Key": "mytag2", "Value": "mytagvalue2"},
        ],
    }
    client.put_bucket_lifecycle_configuration(
        Bucket="bucket", LifecycleConfiguration=lfc
    )
    result = client.get_bucket_lifecycle_configuration(Bucket="bucket")
    assert len(result["Rules"]) == 1
    assert not result["Rules"][0]["Filter"].get("Prefix")
    assert result["Rules"][0]["Filter"]["And"]["Prefix"] == "some/prefix"
    assert len(result["Rules"][0]["Filter"]["And"]["Tags"]) == 2
    assert result["Rules"][0]["Filter"]["And"]["Tags"][0]["Key"] == "mytag"
    assert result["Rules"][0]["Filter"]["And"]["Tags"][0]["Value"] == "mytagvalue"
    assert result["Rules"][0]["Filter"]["And"]["Tags"][1]["Key"] == "mytag2"
    assert result["Rules"][0]["Filter"]["And"]["Tags"][1]["Value"] == "mytagvalue2"
    with pytest.raises(KeyError):
        assert result["Rules"][0]["Prefix"]

    # And filter without Prefix but multiple Tags:
    lfc["Rules"][0]["Filter"]["And"] = {
        "Tags": [
            {"Key": "mytag", "Value": "mytagvalue"},
            {"Key": "mytag2", "Value": "mytagvalue2"},
        ]
    }
    client.put_bucket_lifecycle_configuration(
        Bucket="bucket", LifecycleConfiguration=lfc
    )
    result = client.get_bucket_lifecycle_configuration(Bucket="bucket")
    assert len(result["Rules"]) == 1
    with pytest.raises(KeyError):
        assert result["Rules"][0]["Filter"]["And"]["Prefix"]
    assert len(result["Rules"][0]["Filter"]["And"]["Tags"]) == 2
    assert result["Rules"][0]["Filter"]["And"]["Tags"][0]["Key"] == "mytag"
    assert result["Rules"][0]["Filter"]["And"]["Tags"][0]["Value"] == "mytagvalue"
    assert result["Rules"][0]["Filter"]["And"]["Tags"][1]["Key"] == "mytag2"
    assert result["Rules"][0]["Filter"]["And"]["Tags"][1]["Value"] == "mytagvalue2"
    with pytest.raises(KeyError):
        assert result["Rules"][0]["Prefix"]

    # Can't have both filter and prefix:
    lfc["Rules"][0]["Prefix"] = ""
    with pytest.raises(ClientError) as err:
        client.put_bucket_lifecycle_configuration(
            Bucket="bucket", LifecycleConfiguration=lfc
        )
    assert err.value.response["Error"]["Code"] == "MalformedXML"

    lfc["Rules"][0]["Prefix"] = "some/path"
    with pytest.raises(ClientError) as err:
        client.put_bucket_lifecycle_configuration(
            Bucket="bucket", LifecycleConfiguration=lfc
        )
    assert err.value.response["Error"]["Code"] == "MalformedXML"

    # No filters -- just a prefix:
    del lfc["Rules"][0]["Filter"]
    client.put_bucket_lifecycle_configuration(
        Bucket="bucket", LifecycleConfiguration=lfc
    )
    result = client.get_bucket_lifecycle_configuration(Bucket="bucket")
    assert not result["Rules"][0].get("Filter")
    assert result["Rules"][0]["Prefix"] == "some/path"

    # Can't have Tag, Prefix, and And in a filter:
    del lfc["Rules"][0]["Prefix"]
    lfc["Rules"][0]["Filter"] = {
        "Prefix": "some/prefix",
        "Tag": {"Key": "mytag", "Value": "mytagvalue"},
    }
    with pytest.raises(ClientError) as err:
        client.put_bucket_lifecycle_configuration(
            Bucket="bucket", LifecycleConfiguration=lfc
        )
    assert err.value.response["Error"]["Code"] == "MalformedXML"

    lfc["Rules"][0]["Filter"] = {
        "Tag": {"Key": "mytag", "Value": "mytagvalue"},
        "And": {
            "Prefix": "some/prefix",
            "Tags": [
                {"Key": "mytag", "Value": "mytagvalue"},
                {"Key": "mytag2", "Value": "mytagvalue2"},
            ],
        },
    }
    with pytest.raises(ClientError) as err:
        client.put_bucket_lifecycle_configuration(
            Bucket="bucket", LifecycleConfiguration=lfc
        )
    assert err.value.response["Error"]["Code"] == "MalformedXML"

    # Make sure multiple rules work:
    lfc = {
        "Rules": [
            {
                "Expiration": {"Days": 7},
                "ID": "wholebucket",
                "Filter": {"Prefix": ""},
                "Status": "Enabled",
            },
            {
                "Expiration": {"Days": 10},
                "ID": "Tags",
                "Filter": {"Tag": {"Key": "somekey", "Value": "somevalue"}},
                "Status": "Enabled",
            },
        ]
    }
    client.put_bucket_lifecycle_configuration(
        Bucket="bucket", LifecycleConfiguration=lfc
    )
    result = client.get_bucket_lifecycle_configuration(Bucket="bucket")["Rules"]
    assert len(result) == 2
    assert result[0]["ID"] == "wholebucket"
    assert result[1]["ID"] == "Tags"


@mock_aws
def test_lifecycle_with_eodm():
    client = boto3.client("s3", DEFAULT_REGION_NAME)
    client.create_bucket(Bucket="bucket")

    lfc = {
        "Rules": [
            {
                "Expiration": {"ExpiredObjectDeleteMarker": True},
                "ID": "wholebucket",
                "Filter": {"Prefix": ""},
                "Status": "Enabled",
            }
        ]
    }
    client.put_bucket_lifecycle_configuration(
        Bucket="bucket", LifecycleConfiguration=lfc
    )
    result = client.get_bucket_lifecycle_configuration(Bucket="bucket")
    assert len(result["Rules"]) == 1
    assert result["Rules"][0]["Expiration"]["ExpiredObjectDeleteMarker"]

    # Set to False:
    lfc["Rules"][0]["Expiration"]["ExpiredObjectDeleteMarker"] = False
    client.put_bucket_lifecycle_configuration(
        Bucket="bucket", LifecycleConfiguration=lfc
    )
    result = client.get_bucket_lifecycle_configuration(Bucket="bucket")
    assert len(result["Rules"]) == 1
    assert not result["Rules"][0]["Expiration"]["ExpiredObjectDeleteMarker"]

    # With failure:
    lfc["Rules"][0]["Expiration"]["Days"] = 7
    with pytest.raises(ClientError) as err:
        client.put_bucket_lifecycle_configuration(
            Bucket="bucket", LifecycleConfiguration=lfc
        )
    assert err.value.response["Error"]["Code"] == "MalformedXML"
    del lfc["Rules"][0]["Expiration"]["Days"]

    lfc["Rules"][0]["Expiration"]["Date"] = datetime(2015, 1, 1)
    with pytest.raises(ClientError) as err:
        client.put_bucket_lifecycle_configuration(
            Bucket="bucket", LifecycleConfiguration=lfc
        )
    assert err.value.response["Error"]["Code"] == "MalformedXML"


@mock_aws
def test_lifecycle_with_nve():
    client = boto3.client("s3", DEFAULT_REGION_NAME)
    client.create_bucket(Bucket="bucket")

    lfc = {
        "Rules": [
            {
                "NoncurrentVersionExpiration": {"NoncurrentDays": 30},
                "ID": "wholebucket",
                "Filter": {"Prefix": ""},
                "Status": "Enabled",
            }
        ]
    }
    client.put_bucket_lifecycle_configuration(
        Bucket="bucket", LifecycleConfiguration=lfc
    )
    result = client.get_bucket_lifecycle_configuration(Bucket="bucket")
    assert len(result["Rules"]) == 1
    assert result["Rules"][0]["NoncurrentVersionExpiration"]["NoncurrentDays"] == 30

    # Change NoncurrentDays:
    lfc["Rules"][0]["NoncurrentVersionExpiration"]["NoncurrentDays"] = 10
    client.put_bucket_lifecycle_configuration(
        Bucket="bucket", LifecycleConfiguration=lfc
    )
    result = client.get_bucket_lifecycle_configuration(Bucket="bucket")
    assert len(result["Rules"]) == 1
    assert result["Rules"][0]["NoncurrentVersionExpiration"]["NoncurrentDays"] == 10

    # TODO: Add test for failures due to missing children


@mock_aws
def test_lifecycle_with_nvt():
    client = boto3.client("s3", DEFAULT_REGION_NAME)
    client.create_bucket(Bucket="bucket")

    lfc = {
        "Rules": [
            {
                "NoncurrentVersionTransitions": [
                    {"NoncurrentDays": 30, "StorageClass": "ONEZONE_IA"}
                ],
                "ID": "wholebucket",
                "Filter": {"Prefix": ""},
                "Status": "Enabled",
            }
        ]
    }
    client.put_bucket_lifecycle_configuration(
        Bucket="bucket", LifecycleConfiguration=lfc
    )
    result = client.get_bucket_lifecycle_configuration(Bucket="bucket")
    assert len(result["Rules"]) == 1
    assert result["Rules"][0]["NoncurrentVersionTransitions"][0]["NoncurrentDays"] == 30
    assert (
        result["Rules"][0]["NoncurrentVersionTransitions"][0]["StorageClass"]
        == "ONEZONE_IA"
    )

    # Change NoncurrentDays:
    lfc["Rules"][0]["NoncurrentVersionTransitions"][0]["NoncurrentDays"] = 10
    client.put_bucket_lifecycle_configuration(
        Bucket="bucket", LifecycleConfiguration=lfc
    )
    result = client.get_bucket_lifecycle_configuration(Bucket="bucket")
    assert len(result["Rules"]) == 1
    assert result["Rules"][0]["NoncurrentVersionTransitions"][0]["NoncurrentDays"] == 10

    # Change StorageClass:
    lfc["Rules"][0]["NoncurrentVersionTransitions"][0]["StorageClass"] = "GLACIER"
    client.put_bucket_lifecycle_configuration(
        Bucket="bucket", LifecycleConfiguration=lfc
    )
    result = client.get_bucket_lifecycle_configuration(Bucket="bucket")
    assert len(result["Rules"]) == 1
    assert (
        result["Rules"][0]["NoncurrentVersionTransitions"][0]["StorageClass"]
        == "GLACIER"
    )

    # With failures for missing children:
    del lfc["Rules"][0]["NoncurrentVersionTransitions"][0]["NoncurrentDays"]
    with pytest.raises(ClientError) as err:
        client.put_bucket_lifecycle_configuration(
            Bucket="bucket", LifecycleConfiguration=lfc
        )
    assert err.value.response["Error"]["Code"] == "MalformedXML"
    lfc["Rules"][0]["NoncurrentVersionTransitions"][0]["NoncurrentDays"] = 30

    del lfc["Rules"][0]["NoncurrentVersionTransitions"][0]["StorageClass"]
    with pytest.raises(ClientError) as err:
        client.put_bucket_lifecycle_configuration(
            Bucket="bucket", LifecycleConfiguration=lfc
        )
    assert err.value.response["Error"]["Code"] == "MalformedXML"


@mock_aws
def test_lifecycle_with_multiple_nvt():
    client = boto3.client("s3", DEFAULT_REGION_NAME)
    client.create_bucket(Bucket="bucket")

    lfc = {
        "Rules": [
            {
                "NoncurrentVersionTransitions": [
                    {"NoncurrentDays": 30, "StorageClass": "ONEZONE_IA"},
                    {"NoncurrentDays": 50, "StorageClass": "GLACIER"},
                ],
                "ID": "wholebucket",
                "Filter": {"Prefix": ""},
                "Status": "Enabled",
            }
        ]
    }
    client.put_bucket_lifecycle_configuration(
        Bucket="bucket", LifecycleConfiguration=lfc
    )
    result = client.get_bucket_lifecycle_configuration(Bucket="bucket")
    assert len(result["Rules"]) == 1
    assert result["Rules"][0]["NoncurrentVersionTransitions"][0] == {
        "NoncurrentDays": 30,
        "StorageClass": "ONEZONE_IA",
    }
    assert result["Rules"][0]["NoncurrentVersionTransitions"][1] == {
        "NoncurrentDays": 50,
        "StorageClass": "GLACIER",
    }


@mock_aws
def test_lifecycle_with_multiple_transitions():
    client = boto3.client("s3", DEFAULT_REGION_NAME)
    client.create_bucket(Bucket="bucket")

    lfc = {
        "Rules": [
            {
                "Transitions": [
                    {"Days": 30, "StorageClass": "ONEZONE_IA"},
                    {"Days": 50, "StorageClass": "GLACIER"},
                ],
                "ID": "wholebucket",
                "Filter": {"Prefix": ""},
                "Status": "Enabled",
            }
        ]
    }
    client.put_bucket_lifecycle_configuration(
        Bucket="bucket", LifecycleConfiguration=lfc
    )
    result = client.get_bucket_lifecycle_configuration(Bucket="bucket")
    assert len(result["Rules"]) == 1
    assert result["Rules"][0]["Transitions"][0] == {
        "Days": 30,
        "StorageClass": "ONEZONE_IA",
    }
    assert result["Rules"][0]["Transitions"][1] == {
        "Days": 50,
        "StorageClass": "GLACIER",
    }


@mock_aws
def test_lifecycle_with_aimu():
    client = boto3.client("s3", DEFAULT_REGION_NAME)
    client.create_bucket(Bucket="bucket")

    lfc = {
        "Rules": [
            {
                "AbortIncompleteMultipartUpload": {"DaysAfterInitiation": 7},
                "ID": "wholebucket",
                "Filter": {"Prefix": ""},
                "Status": "Enabled",
            }
        ]
    }
    client.put_bucket_lifecycle_configuration(
        Bucket="bucket", LifecycleConfiguration=lfc
    )
    result = client.get_bucket_lifecycle_configuration(Bucket="bucket")
    assert len(result["Rules"]) == 1
    assert (
        result["Rules"][0]["AbortIncompleteMultipartUpload"]["DaysAfterInitiation"] == 7
    )

    # Change DaysAfterInitiation:
    lfc["Rules"][0]["AbortIncompleteMultipartUpload"]["DaysAfterInitiation"] = 30
    client.put_bucket_lifecycle_configuration(
        Bucket="bucket", LifecycleConfiguration=lfc
    )
    result = client.get_bucket_lifecycle_configuration(Bucket="bucket")
    assert len(result["Rules"]) == 1
    assert (
        result["Rules"][0]["AbortIncompleteMultipartUpload"]["DaysAfterInitiation"]
        == 30
    )

    # TODO: Add test for failures due to missing children


@mock_aws
def test_lifecycle_with_glacier_transition_boto3():
    s3_resource = boto3.resource("s3", region_name="us-east-1")
    client = boto3.client("s3", region_name="us-east-1")
    s3_resource.create_bucket(Bucket="foobar")

    client.put_bucket_lifecycle_configuration(
        Bucket="foobar",
        LifecycleConfiguration={
            "Rules": [
                {
                    "ID": "myid",
                    "Prefix": "",
                    "Status": "Enabled",
                    "Transitions": [{"Days": 30, "StorageClass": "GLACIER"}],
                }
            ]
        },
    )

    response = client.get_bucket_lifecycle_configuration(Bucket="foobar")
    assert "Rules" in response
    rules = response["Rules"]
    assert len(rules) == 1
    assert rules[0]["ID"] == "myid"
    transition = rules[0]["Transitions"][0]
    assert transition["Days"] == 30
    assert transition["StorageClass"] == "GLACIER"
    assert "Date" not in transition


@mock_aws
def test_lifecycle_multi_boto3():
    s3_resource = boto3.resource("s3", region_name="us-east-1")
    client = boto3.client("s3", region_name="us-east-1")
    s3_resource.create_bucket(Bucket="foobar")

    date = "2022-10-12T00:00:00.000Z"
    storage_class = "GLACIER"

    client.put_bucket_lifecycle_configuration(
        Bucket="foobar",
        LifecycleConfiguration={
            "Rules": [
                {"ID": "1", "Prefix": "1/", "Status": "Enabled"},
                {
                    "ID": "2",
                    "Prefix": "2/",
                    "Status": "Enabled",
                    "Expiration": {"Days": 2},
                },
                {
                    "ID": "3",
                    "Prefix": "3/",
                    "Status": "Enabled",
                    "Expiration": {"Date": date},
                },
                {
                    "ID": "4",
                    "Prefix": "4/",
                    "Status": "Enabled",
                    "Transitions": [{"Days": 4, "StorageClass": "GLACIER"}],
                },
                {
                    "ID": "5",
                    "Prefix": "5/",
                    "Status": "Enabled",
                    "Transitions": [{"Date": date, "StorageClass": "GLACIER"}],
                },
            ]
        },
    )

    # read the lifecycle back
    rules = client.get_bucket_lifecycle_configuration(Bucket="foobar")["Rules"]

    for rule in rules:
        if rule["ID"] == "1":
            assert rule["Prefix"] == "1/"
            assert "Expiration" not in rule
        elif rule["ID"] == "2":
            assert rule["Prefix"] == "2/"
            assert rule["Expiration"]["Days"] == 2
        elif rule["ID"] == "3":
            assert rule["Prefix"] == "3/"
            assert isinstance(rule["Expiration"]["Date"], datetime)
            assert rule["Expiration"]["Date"].strftime("%Y-%m-%dT%H:%M:%S.000Z") == date
        elif rule["ID"] == "4":
            assert rule["Prefix"] == "4/"
            assert rule["Transitions"][0]["Days"] == 4
            assert rule["Transitions"][0]["StorageClass"] == storage_class
        elif rule["ID"] == "5":
            assert rule["Prefix"] == "5/"
            assert isinstance(rule["Transitions"][0]["Date"], datetime)
            assert (
                rule["Transitions"][0]["Date"].strftime("%Y-%m-%dT%H:%M:%S.000Z")
                == date
            )
            assert rule["Transitions"][0]["StorageClass"] == storage_class
        else:
            assert False, "Invalid rule id"


@mock_aws
def test_lifecycle_delete_boto3():
    s3_resource = boto3.resource("s3", region_name="us-east-1")
    client = boto3.client("s3", region_name="us-east-1")
    s3_resource.create_bucket(Bucket="foobar")

    client.put_bucket_lifecycle_configuration(
        Bucket="foobar",
        LifecycleConfiguration={
            "Rules": [{"ID": "myid", "Prefix": "", "Status": "Enabled"}]
        },
    )

    client.delete_bucket_lifecycle(Bucket="foobar")

    with pytest.raises(ClientError) as ex:
        client.get_bucket_lifecycle_configuration(Bucket="foobar")
    assert ex.value.response["Error"]["Code"] == "NoSuchLifecycleConfiguration"
    assert ex.value.response["Error"]["Message"] == (
        "The lifecycle configuration does not exist"
    )



"""Exceptions raised by the directconnect service."""

from moto.core.exceptions import JsonRESTError


class DXConnectionError(JsonRESTError):
    code = 400


class ConnectionIdMissing(DXConnectionError):
    def __init__(self) -> None:
        super().__init__("ConnectionIdMissing", "The connection ID is missing.")


class ConnectionNotFound(DXConnectionError):
    def __init__(self, connection_id: str, region: str) -> None:
        super().__init__(
            "ConnectionNotFound",
            f"{connection_id} does not match any connections in region {region}.",
        )


class LAGNotFound(DXConnectionError):
    def __init__(self, lag_id: str, region: str) -> None:
        super().__init__(
            "LagNotFound",
            f"{lag_id} does not match any LAGs in region {region}.",
        )


class MacSecKeyNotFound(DXConnectionError):
    def __init__(self, secret_arn: str, connection_id: str) -> None:
        super().__init__(
            "MacSecKeyNotFound",
            f"Secret ARN '{secret_arn}' not found for connection {connection_id}.",
        )



"""inspector2 base URL and path."""

from .responses import Inspector2Response

url_bases = [
    r"https?://inspector2\.(.+)\.amazonaws\.com",
]


url_paths = {
    "{0}/delegatedadminaccounts/disable$": Inspector2Response.dispatch,
    "{0}/delegatedadminaccounts/enable$": Inspector2Response.dispatch,
    "{0}/delegatedadminaccounts/list$": Inspector2Response.dispatch,
    "{0}/disable$": Inspector2Response.dispatch,
    "{0}/enable$": Inspector2Response.dispatch,
    "{0}/filters/create$": Inspector2Response.dispatch,
    "{0}/filters/delete$": Inspector2Response.dispatch,
    "{0}/filters/list$": Inspector2Response.dispatch,
    "{0}/findings/list$": Inspector2Response.dispatch,
    "{0}/members/associate$": Inspector2Response.dispatch,
    "{0}/members/get": Inspector2Response.dispatch,
    "{0}/members/list": Inspector2Response.dispatch,
    "{0}/members/disassociate$": Inspector2Response.dispatch,
    "{0}/status/batch/get$": Inspector2Response.dispatch,
    "{0}/organizationconfiguration/describe$": Inspector2Response.dispatch,
    "{0}/organizationconfiguration/update$": Inspector2Response.dispatch,
    "{0}/tags/(?P<resource_arn>.+)$": Inspector2Response.dispatch,
}



import abc
from typing import Final, Optional

from moto.stepfunctions.parser.api import ExecutionFailedEventDetails, Timestamp


class ProgramState(abc.ABC): ...


class ProgramEnded(ProgramState):
    pass


class ProgramStopped(ProgramState):
    def __init__(
        self, stop_date: Timestamp, error: Optional[str], cause: Optional[str]
    ):
        super().__init__()
        self.stop_date: Timestamp = stop_date
        self.error: Optional[str] = error
        self.cause: Optional[str] = cause


class ProgramRunning(ProgramState):
    def __init__(self):
        super().__init__()
        self._next_state_name: Optional[str] = None

    @property
    def next_state_name(self) -> str:
        next_state_name = self._next_state_name
        if next_state_name is None:
            raise RuntimeError(
                "Could not retrieve NextState from uninitialised ProgramState."
            )
        return next_state_name

    @next_state_name.setter
    def next_state_name(self, next_state_name) -> None:
        if not self._validate_next_state_name(next_state_name):
            raise ValueError(f"No such NextState '{next_state_name}'.")
        self._next_state_name = next_state_name

    @staticmethod
    def _validate_next_state_name(next_state_name: Optional[str]) -> bool:
        # TODO.
        return bool(next_state_name)


class ProgramError(ProgramState):
    error: Final[Optional[ExecutionFailedEventDetails]]

    def __init__(self, error: Optional[ExecutionFailedEventDetails]):
        super().__init__()
        self.error = error


class ProgramTimedOut(ProgramState):
    pass



from moto.core import DEFAULT_ACCOUNT_ID as ACCOUNT_ID

TEST_DESCRIPTION = "test_description"

TEST_TAGS = {"key1": "value1", "key2": "value2"}

TEST_AVAILABLE_STATUS = "AVAILABLE"
TEST_DELETING_STATUS = "DELETING"

TEST_REGISTRY_NAME = "TestRegistry"
TEST_INVALID_REGISTRY_NAME_DOES_NOT_EXIST = "InvalidRegistryDoesNotExist"
TEST_REGISTRY_ARN = f"arn:aws:glue:us-east-1:{ACCOUNT_ID}:registry/{TEST_REGISTRY_NAME}"
TEST_REGISTRY_ID = {"RegistryName": f"{TEST_REGISTRY_NAME}"}

TEST_SCHEMA_NAME = "TestSchema"
TEST_INVALID_SCHEMA_NAME_DOES_NOT_EXIST = "InvalidSchemaDoesNotExist"
TEST_SCHEMA_ARN = f"arn:aws:glue:us-east-1:{ACCOUNT_ID}:schema/{TEST_REGISTRY_NAME}/{TEST_SCHEMA_NAME}"
TEST_SCHEMA_ID = {
    "RegistryName": f"{TEST_REGISTRY_NAME}",
    "SchemaName": f"{TEST_SCHEMA_NAME}",
}
TEST_INVALID_SCHEMA_ID_SCHEMA_DOES_NOT_EXIST = {
    "RegistryName": TEST_REGISTRY_NAME,
    "SchemaName": TEST_INVALID_SCHEMA_NAME_DOES_NOT_EXIST,
}
TEST_INVALID_SCHEMA_ID_REGISTRY_DOES_NOT_EXIST = {
    "RegistryName": TEST_INVALID_REGISTRY_NAME_DOES_NOT_EXIST,
    "SchemaName": TEST_SCHEMA_NAME,
}

TEST_AVRO_DATA_FORMAT = "AVRO"
TEST_JSON_DATA_FORMAT = "JSON"
TEST_PROTOBUF_DATA_FORMAT = "PROTOBUF"

TEST_BACKWARD_COMPATIBILITY = "BACKWARD"
TEST_DISABLED_COMPATIBILITY = "DISABLED"

TEST_SCHEMA_VERSION_NUMBER = {"VersionNumber": 1, "LatestVersion": False}
TEST_SCHEMA_VERSION_NUMBER_LATEST_VERSION = {"LatestVersion": True}
TEST_VERSION_ID = "00000000-0000-0000-0000-000000000000"


TEST_METADATA_KEY = "test_metadata_key"
TEST_METADATA_VALUE = "test_metadata_value"
TEST_METADATA_KEY_VALUE = {
    "MetadataKey": TEST_METADATA_KEY,
    "MetadataValue": TEST_METADATA_VALUE,
}

TEST_AVRO_SCHEMA_DEFINITION = """{
        "type": "record",
        "namespace": "Moto_Test",
        "name": "Person",
        "fields": [
            {
                "name": "Name",
                "type": "string"
            },
            {
                "name": "Age",
                "type": "int"
            }
        ]
    }"""

TEST_NEW_AVRO_SCHEMA_DEFINITION = """{
        "type": "record",
        "namespace": "Moto_Test",
        "name": "Person",
        "fields": [
            {
                "name": "Name",
                "type": "string"
            },
            {
                "name": "Age",
                "type": "int"
            },
            {
                "name": "address",
                "type": "string",
                "default": ""
            }
        ]
    }"""

TEST_JSON_SCHEMA_DEFINITION = """{
        "$id": "https://example.com/person.schema.json",
        "$schema": "http://json-schema.org/draft-07/schema#",
        "title": "Person",
        "type": "object",
        "properties": {
            "firstName": {
                "type": "string",
                "description": "The person's first name."
            },
            "lastName": {
                "type": "string",
                "description": "The person's last name."
            }
        }
    }"""

TEST_NEW_JSON_SCHEMA_DEFINITION = """{
        "$id": "https://example.com/person.schema.json",
        "$schema": "http://json-schema.org/draft-07/schema#",
        "title": "Person",
        "type": "object",
        "properties": {
            "firstName": {
                "type": "string",
                "description": "The person's first name."
            },
            "lastName": {
                "type": "string",
                "description": "The person's last name."
            },
            "age": {
                "description": "Age in years which must be equal to or greater than zero.",
                "type": "integer",
                "minimum": 0
            }
        }
    }"""

TEST_PROTOBUF_SCHEMA_DEFINITION = """syntax = "proto2";
    package tutorial;

    option java_multiple_files = true;
    option java_package = "com.example.tutorial.protos";
    option java_outer_classname = "AddressBookProtos";

    message Person {
      optional string name = 1;
      optional int32 id = 2;
      optional string email = 3;

      enum PhoneType {
        MOBILE = 0;
        HOME = 1;
        WORK = 2;
      }

      message PhoneNumber {
        optional string number = 1;
        optional PhoneType type = 2 [default = HOME];
      }

      repeated PhoneNumber phones = 4;
    }"""

TEST_NEW_PROTOBUF_SCHEMA_DEFINITION = """syntax = "proto2";

    package tutorial;

    option java_multiple_files = true;
    option java_package = "com.example.tutorial.protos";
    option java_outer_classname = "AddressBookProtos";

    message Person {
      optional string name = 1;
      optional int32 id = 2;
      optional string email = 3;

      enum PhoneType {
        MOBILE = 0;
        HOME = 1;
        WORK = 2;
      }

      message PhoneNumber {
        optional string number = 1;
        optional PhoneType type = 2 [default = HOME];
      }

      repeated PhoneNumber phones = 4;
    }

    message AddressBook {
      repeated Person people = 1;
    }"""
